"""
The primitives module generates basic 3D objects

**Conventions:**

arrays follow the ZXY convention, with

    - Z : depth axis (axial, focus axis)
    - X : horizontal axis (lateral)
    - Y : vertical axis (lateral, rotation axis when relevant)
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import skimage.transform


def quadrant_symmetry(quadrant):
    """
    Generate a binary quadrant by rotating itself 90° on all 3 axes and summing.

    Parameters
    ----------
    quadrant : numpy.ndarray(bool)
        The original 3D quadrant.

    Returns
    -------
    numpy.ndarray
        The symmetrized quadrant. Will have cubic shape equal to the largest
        dimension of the original quadrant.
    """

    size = np.max(quadrant.shape)

    full_quadrant = np.zeros((size, size, size), dtype=bool)
    full_quadrant[:quadrant.shape[0], :quadrant.shape[1],
                  :quadrant.shape[2]] = quadrant
    full_quadrant[:quadrant.shape[1], :quadrant.shape[2],
                  :quadrant.shape[0]] |= quadrant.transpose((1, 2, 0))
    full_quadrant[:quadrant.shape[2], :quadrant.shape[0],
                  :quadrant.shape[1]] |= quadrant.transpose((2, 0, 1))

    return full_quadrant


def quadrant_to_volume(quadrant, odd=(False, False, False)):
    """
    Generate a volume by mirroring a quadrant in all 8 corners.

    Parameters
    ----------
    quadrant : numpy.ndarray
        The quadrant corresponding to the end of the axes.
    odd : tuple, optional
        If the target dimensions are odd, by default (False, False, False).
        If even, the dimension will be ``2 * quandrant.shape``.
        If odd, the dimension will be ``2 * quadrant.shape - 1``.

    Returns
    -------
    numpy.ndarray
        The full volume.
    """

    volume = np.empty(
        (2 * quadrant.shape[0] - odd[0], 2 * quadrant.shape[1] - odd[1], 2 * quadrant.shape[2] - odd[2]), dtype=quadrant.dtype)
    volume[quadrant.shape[0] - odd[0]:, quadrant.shape[1] -
           odd[1]:, quadrant.shape[2] - odd[2]:] = quadrant
    volume[:quadrant.shape[0], quadrant.shape[1] - odd[1]:,
           quadrant.shape[2] - odd[2]:] = np.flip(quadrant, 0)
    volume[:, :quadrant.shape[1] - odd[1], quadrant.shape[2] - odd[2]:] = np.flip(
        volume[:, quadrant.shape[1]:, quadrant.shape[2] - odd[2]:], 1)
    volume[:, :, :quadrant.shape[2] - odd[2]] = np.flip(
        volume[:, :, quadrant.shape[2]:], 2)

    return volume


def boccia(size, radius=None, n_stripes=3, deg_space=15, deg_width=7.5, rad_thick=0.12,
           antialias=2, dtype=np.float64):
    """
    Create a boccia simulated sample: resolution stripes on a sphere

    Parameters
    ----------
    size : int
        side of the cube containing the volume (must be even)
    radius : float, optional
        radius of the boccia, by default None (will be size-1)
    n_stripes : int, optional
        number of stripes to generate, by default 3
    deg_space : int, optional
        spacing in degrees between the center of the stripes, by default 15
    deg_width : float, optional
        width in degrees of the stripes, by default 7.5
    rad_thick : float, optional
        thickness of the stripes, as a proportion of the radius, by default 0.12
    antialias : int, optional
        antialiasing scale factor, by default 2
    dtype : numpy.dtype or str, optional
        the datatype, by default numpy.float64

    Returns
    -------
    numpy.ndarray
        The volume containing the boccia.

    Raises
    ------
    ValueError
        if the size is odd
    ValueError
        if the antialias is negative
    ValueError
        if the angle surpasses 90°
    """

    if size % 2:
        raise ValueError('The size must be even to cut it in quadrants')

    if antialias < 1 or not isinstance(antialias, int):
        raise ValueError('Antialias must be a positive integer')

    size //= 2
    if radius is None:
        radius = size - 1  # leave one pixel out for antialiasing
    size *= antialias
    radius *= antialias

    half_width = np.deg2rad(deg_width) / 2

    # the angles at the center of the stripes
    c_angles = np.arange(n_stripes / 2)

    # if even number of stripes
    if not n_stripes % 2:
        c_angles += 0.5

    c_angles *= np.deg2rad(deg_space)

    max_angle = c_angles[-1] + half_width
    if max_angle > np.pi / 2:
        raise ValueError('max angle must be less than 90°')

    width = int(np.ceil(size * np.sin(max_angle)))

    x = (np.arange(size) + 0.5).reshape((size, 1, 1))
    y = (np.arange(size) + 0.5).reshape((1, size, 1))
    z = (np.arange(width) + 0.5).reshape((1, 1, width))

    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    phi = np.arccos(z / r)
    phi = np.pi / 2 - phi

    angle_crit = np.zeros_like(phi, dtype=bool)
    for angle in c_angles:
        angle_crit |= (phi > (angle - half_width)
                       ) & (phi < (angle + half_width))

    quadrant = (r < radius) & (r > (radius * (1 - rad_thick))) & angle_crit

    quadrant = quadrant_symmetry(quadrant).astype(dtype)

    if antialias > 1:
        quadrant = skimage.transform.rescale(
            quadrant, 1 / antialias, mode='symmetric')

    return quadrant_to_volume(quadrant)


def torus_boccia(size, radius=None, n_stripes=3, deg_space=15, torus_radius=0.075,
                 antialias=2, dtype=np.float64):
    """
    Generate a boccia with torus stripes

    Parameters
    ----------
    size : int
        side of the cube containing the volume (must be even)
    radius : float, optional
        radius of the boccia, by default None
    n_stripes : int, optional
        number of torus stripes, by default 3
    deg_space : int, optional
        spacing of the torus on the sphere in degrees, by default 15
    torus_radius : float, optional
        radius of the torus, as a fraction of the sphere radius, by default 0.075
    antialias : int, optional
        antialiasing scale factor, by default 2
    dtype : numpy.dtype or str, optional
        the datatype, by default numpy.float64

    Returns
    -------
    numpy.ndarray
        The volume containing the torus boccia.

    Raises
    ------
    ValueError
        if the size is odd
    ValueError
        if the antialias is negative
    ValueError
        if the angle surpasses 90°
    """

    if size % 2:
        raise ValueError('The size must be even to cut it in quadrants')

    if antialias < 1 or not isinstance(antialias, int):
        raise ValueError('Antialias must be a positive integer')

    size //= 2
    if radius is None:
        radius = size - 1  # leave one pixel out for antialiasing
    size *= antialias
    radius *= antialias

    # the angles at the center of the stripes
    c_angles = np.arange(n_stripes / 2)

    # if even number of stripes
    if not n_stripes % 2:
        c_angles += 0.5

    c_angles *= np.deg2rad(deg_space)

    max_angle = c_angles[-1] + np.arcsin(torus_radius)
    if max_angle > np.pi / 2:
        raise ValueError('max angle must be less than 90°')

    c_pos = np.array([np.cos(c_angles), np.sin(c_angles)]) * \
        radius * (1 - torus_radius)
    width = int(np.ceil(c_pos[1, :].max() + size * torus_radius))

    x = (np.arange(size) + 0.5).reshape((size, 1, 1, 1))
    y = (np.arange(size) + 0.5).reshape((1, size, 1, 1))
    z = (np.arange(width) + 0.5).reshape((1, 1, width, 1))

    r = np.sqrt(x ** 2 + y ** 2)

    rad = (r - c_pos[0, :]) ** 2 + (z - c_pos[1, :]) ** 2
    rad = rad.min(-1)

    quadrant = rad < (torus_radius * radius) ** 2

    quadrant = quadrant_symmetry(quadrant).astype(dtype)

    if antialias > 1:
        quadrant = skimage.transform.rescale(
            quadrant, 1 / antialias, mode='symmetric')

    return quadrant_to_volume(quadrant)


def ball(size, radius=None, in_radius=0, antialias=2, dtype=np.float64):
    """
    Generate hollow ball.

    :param size: 
    :param radius: 
    :param in_radius: inner radius fraction (set to 0 for full ball)
    :param antialias: antialiasing scale factor

    Parameters
    ----------
    size : int
        side of the cube containing the volume
    radius : float, optional
        radius of the ball, by default None
    in_radius : float, optional
        radius of the inner (empty) sphere expressed as a fraction of the outer
        radius [0-1], by default 0 (plain ball)
    antialias : int, optional
        antialiasing scale factor, by default 2
    dtype : numpy.dtype or str, optional
        the datatype, by default numpy.float64

    Returns
    -------
    numpy.ndarray
        The volume containing the ball.

    Raises
    ------
    ValueError
        if the size is odd
    ValueError
        if the antialias is negative
    """

    if size % 2:
        raise ValueError('The size must be even to cut it in quadrants')

    if antialias < 1 or not isinstance(antialias, int):
        raise ValueError('Antialias must be a positive integer')

    size //= 2
    if radius is None:
        radius = size - 1  # leave one pixel out for antialiasing
    size *= antialias
    radius *= antialias

    x = (np.arange(size) + 0.5).reshape((size, 1, 1))
    y = (np.arange(size) + 0.5).reshape((1, size, 1))
    z = (np.arange(size) + 0.5).reshape((1, 1, size))

    r = x ** 2 + y ** 2 + z ** 2

    quadrant = r < radius ** 2

    if in_radius > 0:
        quadrant &= r > (in_radius * radius) ** 2

    quadrant = quadrant.astype(dtype)

    if antialias > 1:
        quadrant = skimage.transform.rescale(
            quadrant, 1 / antialias, mode='symmetric')

    return quadrant_to_volume(quadrant)


def phantom(size, scale=1, antialias=2):
    """
    Generate 3D modified Shepp-Logann phantoms

    Parameters
    ----------
    size : int
        size of the output
    scale : float
        scaling factor for the size of the phantoms
    antialias : int, optional
        antialiasing factor, by default 2

    Returns
    -------
    array [ZXY]
        the phantoms

    Raises
    ------
    ValueError
        if antialias is not a positive integer
    """

    if antialias < 1 or not isinstance(antialias, int):
        raise ValueError('Antialias must be a positive integer')

    #      A         a     b      c     x0      y0     z0   phi  theta   psi
    ellipses = [
        [1,      .6900, .920,  .810,     0,      0,     0,    0,     0,    0],
        [-.8,    .6624, .874,  .780,     0, -.0184,     0,    0,     0,    0],
        [-.2,    .1100, .310,  .220,   .22,      0,     0,  -18,     0,   10],
        [-.2,    .1600, .410,  .280,  -.22,      0,     0,   18,     0,   10],
        [.1,     .2100, .250,  .410,     0,    .35,  -.15,    0,     0,    0],
        [.1,     .0460, .046,  .050,     0,     .1,   .25,    0,     0,    0],
        [.1,     .0460, .046,  .050,     0,    -.1,   .25,    0,     0,    0],
        [.1,     .0460, .023,  .050,  -.08,  -.605,     0,    0,     0,    0],
        [.1,     .0230, .023,  .020,     0,  -.606,     0,    0,     0,    0],
        [.1,     .0230, .046,  .020,   .06,  -.605,     0,    0,     0,    0]
    ]

    size *= antialias

    full_mat = np.zeros((size, size, size), dtype=np.float64)

    for quad in range(8):
        if quad == 0:
            mat = full_mat[:size//2, :size//2, :size//2]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
        elif quad == 1:
            mat = full_mat[size//2:, :size//2, :size//2]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[0] += size//2 * 2 / (size - 1)
        elif quad == 2:
            mat = full_mat[size//2:, size//2:, :size//2]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[0] += size//2 * 2 / (size - 1)
            coords[1] += size//2 * 2 / (size - 1)
        elif quad == 3:
            mat = full_mat[size//2:, size//2:, size//2:]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[0] += size//2 * 2 / (size - 1)
            coords[1] += size//2 * 2 / (size - 1)
            coords[2] += size//2 * 2 / (size - 1)
        elif quad == 4:
            mat = full_mat[:size//2, size//2:, size//2:]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[1] += size//2 * 2 / (size - 1)
            coords[2] += size//2 * 2 / (size - 1)
        elif quad == 5:
            mat = full_mat[:size//2:, :size//2, size//2:]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[2] += size//2 * 2 / (size - 1)
        elif quad == 6:
            mat = full_mat[size//2:, :size//2, size//2:]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[0] += size//2 * 2 / (size - 1)
            coords[2] += size//2 * 2 / (size - 1)
        elif quad == 7:
            mat = full_mat[:size//2, size//2:, :size//2]
            coords = 2 * np.indices(mat.shape,
                                    dtype=np.float64) / (size - 1) - 1
            coords[1] += size//2 * 2 / (size - 1)

        coords /= scale

        for ellipse in ellipses:
            A = ellipse[0]
            a2 = ellipse[1] ** 2
            b2 = ellipse[2] ** 2
            c2 = ellipse[3] ** 2
            x0 = ellipse[4]
            y0 = ellipse[5]
            z0 = ellipse[6]
            phi = ellipse[7] * np.pi / 180
            theta = ellipse[8] * np.pi / 180
            psi = ellipse[9] * np.pi / 180

            cphi = np.cos(phi)
            sphi = np.sin(phi)
            ctheta = np.cos(theta)
            stheta = np.sin(theta)
            cpsi = np.cos(psi)
            spsi = np.sin(psi)

            # Euler rotation matrix with ZXY convention
            rotmat = np.array([
                [
                    ctheta,
                    stheta * sphi,
                    -stheta * cphi,
                ],
                [
                    spsi * stheta,
                    cpsi * cphi - ctheta * sphi * spsi,
                    cpsi * sphi + ctheta * cphi * spsi,
                ],
                [
                    cpsi * stheta,
                    -spsi * cphi - ctheta * sphi * cpsi,
                    -spsi * sphi + ctheta * cphi * cpsi,
                ],
            ])

            rcoords = np.tensordot(rotmat, coords, 1)
            idx = ((rcoords[1, :] - x0) ** 2.0 / a2 +
                   (rcoords[2, :] - y0) ** 2.0 / b2 +
                   (rcoords[0, :] - z0) ** 2.0 / c2 <= 1)

            mat[idx] += A

    if antialias > 1:
        full_mat = skimage.transform.rescale(
            full_mat, 1 / antialias, mode='constant')

    return full_mat


if __name__ == '__main__':
    import napari

    TEST_SIZE = 128

    s_ball = ball(TEST_SIZE)
    s_torus = torus_boccia(TEST_SIZE)
    s_boccia = boccia(TEST_SIZE)
    s_phantom = phantom(TEST_SIZE)

    with napari.gui_qt():
        viewer = napari.view_image(s_ball)
        viewer.add_image(s_torus)
        viewer.add_image(s_boccia)
        viewer.add_image(s_phantom)



"""
The optics module provides simulations of the optics of imaging systems for microscopy

**Conventions:**

arrays follow the ZXY convention, with

    - Z : depth axis (axial, focus axis)
    - X : horizontal axis (lateral)
    - Y : vertical axis (lateral, rotation axis when relevant)
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import math
import astropy.units as u
import numpy as np
import poppy
import scipy.interpolate
from cbi_toolbox.simu import primitives
from cbi_toolbox import utils


def create_wf_1d(wf_object, upsampling=1, scale=1, copy=False):
    """
    Create a 1D wavefront object from an existing wavefront

    Parameters
    ----------
    wf_object : poppy.FresnelWavefront
        the original wavefront
    upsampling : int, optional
        upsampling factor (does not change the field of view), by default 1
    scale : int, optional
        zoom factor (changes the field of view), by default 1
    copy : bool, optional
        return a new object, by default False

    Returns
    -------
    poppy.FresnelWavefront
        a 1D wavefront full of 1 with same properties as the input
    """

    if copy:
        wf_object = wf_object.copy()

    wf = np.ones(
        (1, int(wf_object.shape[1] * upsampling)), dtype=wf_object.wavefront.dtype)
    y, x = np.indices(wf.shape, dtype=float)
    x -= wf.shape[1] / 2

    wf_object._x = x
    wf_object._y = y
    wf_object.wavefront = wf
    wf_object.pixelscale = wf_object.pixelscale / upsampling * scale
    wf_object.n = wf.shape[1]

    return wf_object


def wf_to_2d(wf_object, npix=None, copy=False):
    """
    Convert a 1D wavefront to 2D (for plotting only)

    Parameters
    ----------
    wf_object : poppy.FresnelWavefront
        the 1D wavefront
    npix : int, optional
        crop to a size of npix, by default None
    copy : bool, optional
        return a new object, by default False

    Returns
    -------
    poppy.FresnelWavefront
        the 2D wavefront
    """

    if copy:
        wf_object = wf_object.copy()

    if npix is None:
        size = wf_object.shape[1]
    else:
        size = npix
    center = wf_object.shape[1] // 2
    hw = size // 2

    new_wf = np.zeros_like(wf_object.wavefront, shape=(size, size))

    new_wf[hw, :] = wf_object.wavefront[:, center - hw:center + hw]
    wf_object.wavefront = new_wf

    wf_object._y, wf_object._x = np.indices(wf_object.shape, dtype=float)
    wf_object._y -= wf_object.shape[0] / 2.0
    wf_object._x -= wf_object.shape[0] / 2.0

    return wf_object


def wf_mix(wf1, wf2, ref=None):
    """
    Compute a 2D wavefront by multiplying 2 1D wavefronts (for separable propagation)

    Parameters
    ----------
    wf1 : poppy.FresnelWavefront
        a 1D wavefront
    wf2 : poppy.FresnelWavefront
        a 1D wavefront
    ref : poppy.FresnelWavefront, optional
        reference wavefront for the parameters of the output, by default None (wf1 will be used)

    Returns
    -------
    poppy.FresnelWavefront
        the 2D mixed wavefront

    Raises
    ------
    ValueError
        if the input wavefronts have different pixelscales
    """

    if wf1.pixelscale != wf2.pixelscale:
        raise ValueError("The pixelscale of the input wavefronts must match")

    wfa = wf1.wavefront.squeeze()
    wfb = wf2.wavefront.squeeze()

    mix = np.outer(wfb, wfa)

    if ref is None:
        wf_m = wf1.copy()
    else:
        wf_m = ref.copy()

    wf_m.wavefront = mix

    return wf_m


def resample_wavefront(wf, pixelscale, npixels):
    """
    Resample 1D wavefront to new pixelscale
    (adapted from poppy.poppy_core._resample_wavefront_pixelscale)

    Parameters
    ----------
    wf : poppy.FresnelWavefront
        a 1D wavefront
    pixelscale : astropy.units.[distance] / astropy.units.pixel
        target pixelscale
    npixels : int
        target size in pixels

    Returns
    -------
    poppy.FresnelWavefront
        resampled and resized 1D wavefront
    """

    pixscale_ratio = (wf.pixelscale / pixelscale).decompose().value

    def make_axis(npix, step):
        """ Helper function to make coordinate axis for interpolation """
        return step * (np.arange(-npix // 2, npix // 2, dtype=np.float64))

    # Input and output axes for interpolation.  The interpolated wavefront will be evaluated
    # directly onto the detector axis, so don't need to crop afterwards.
    x_in = make_axis(wf.shape[1], wf.pixelscale.to(u.m / u.pix).value)
    x_out = make_axis(npixels.value, pixelscale.to(u.m / u.pix).value)

    def interpolator(arr):
        """
        Bind arguments to scipy's RectBivariateSpline function.
        For data on a regular 2D grid, RectBivariateSpline is more efficient than interp2d.
        """
        return scipy.interpolate.interp1d(
            x_in, arr, kind='slinear', copy=False, fill_value=0,
            assume_sorted=True, bounds_error=False)

    # Interpolate real and imaginary parts separately
    real_resampled = interpolator(wf.wavefront.real)(x_out)
    imag_resampled = interpolator(wf.wavefront.imag)(x_out)
    new_wf = real_resampled + 1j * imag_resampled

    # enforce conservation of energy:
    new_wf *= 1. / pixscale_ratio

    wf.ispadded = False  # if a pupil detector, avoid auto-cropping padded pixels on output
    wf.wavefront = new_wf
    wf.pixelscale = pixelscale


def openspim_illumination(wavelength=500e-9, refr_index=1.333, laser_radius=1.2e-3,
                          objective_na=0.3, objective_focal=18e-3, slit_opening=10e-3,
                          pixelscale=635e-9, npix_fov=512, rel_thresh=None,
                          simu_size=2048, oversample=16):
    """
    Compute the illumination function of an OpenSPIM device

    Parameters
    ----------
    wavelength : float, optional
        illumination wavelength in meters, by default 500e-9
    refr_index : float, optional
        imaging medium refraction index, by default 1.333
    laser_radius : float, optional
        source laser radius in meters, by default 1.2e-3
    objective_na : float, optional
        illumination objective NA, by default 0.3
    objective_focal : float, optional
        illumination objective focal length in meters, by default 18e-3
    slit_opening : float, optional
        vertical slit opening in meters, by default 10e-3
    pixelscale : float, optional
        target pixelscale in meters per pixel, by default 1.3e-3/2048
    npix_fov : int, optional
        target size in pixels, by default 512
    rel_thresh: float, optional
        relative threshold to crop the beam thickness
        if a full row is below this theshold, all rows after are removed
        will be computed as compared to the maximum pixel
    simu_size : int, optional
        size of the arrays used for simulation, by default 2048
    oversample : int, optional
        oversampling used for the simulation (must be increased sith simu_size), by default 16

    Returns
    -------
    array [ZXY]
        the illumination function
    """

    pixel_width = 1
    wavelength *= u.m
    laser_radius *= u.m
    objective_focal *= u.m
    pixelscale *= (u.m / u.pixel)
    slit_opening *= u.m

    noop = poppy.ScalarTransmission()
    beam_ratio = 1 / oversample

    fov_pixels = npix_fov * u.pixel
    detector = poppy.FresnelOpticalSystem()
    detector.add_detector(fov_pixels=fov_pixels, pixelscale=pixelscale)

    # We approximate the objective aperture with a square one to make it separable
    # Given the shape of the wavefront, we estimate the generated error to be negligible
    objective_radius = math.tan(
        math.asin(objective_na / refr_index)) * objective_focal
    objective_aperture = poppy.RectangleAperture(name='objective aperture',
                                                 width=2 * objective_radius,
                                                 height=2 * objective_radius)
    objective_lens = poppy.QuadraticLens(
        f_lens=objective_focal, name='objective lens')

    obj_aperture = poppy.FresnelOpticalSystem()
    obj_aperture.add_optic(objective_aperture, objective_focal)

    # Implement the objective lens separately to be able to account for refractive index change
    obj_lens = poppy.FresnelOpticalSystem()
    obj_lens.add_optic(objective_lens)

    # Computed as following: going through T1 then CLens then T2
    # is equivalent to going through CLens with focal/4
    # Then the radius is computed as the Fourier transform of the input beam, per 2F lens system
    w0_y = (12.5e-3 * u.m * wavelength) / (2 * np.pi ** 2 * laser_radius)
    laser_shape_y = poppy.GaussianAperture(w=w0_y, pupil_diam=5 * w0_y)
    path_y = poppy.FresnelOpticalSystem(
        pupil_diameter=2 * w0_y, npix=pixel_width, beam_ratio=beam_ratio)
    path_y.add_optic(laser_shape_y)

    # Going through T1, slit and T2 is equivalent to going through a half-sized slit,
    # then propagating 1/4 the distance
    # Since we use 1D propagation, we can increase oversampling a lot for better results
    laser_shape_z = poppy.GaussianAperture(
        w=laser_radius, pupil_diam=slit_opening / 2)
    slit = poppy.RectangleAperture(
        name='Slit', width=slit_opening / 2, height=slit_opening / 2)
    path_z = poppy.FresnelOpticalSystem(
        pupil_diameter=slit_opening / 2, npix=pixel_width, beam_ratio=beam_ratio)
    path_z.add_optic(laser_shape_z)
    path_z.add_optic(slit)
    path_z.add_optic(noop, 0.25 * 100e-3 * u.m)

    # Propagate 1D signals
    wf_z = path_z.input_wavefront(wavelength=wavelength)
    create_wf_1d(wf_z, upsampling=simu_size)
    path_z.propagate(wf_z)

    wf_y = path_y.input_wavefront(wavelength=wavelength)
    create_wf_1d(wf_y, upsampling=simu_size, scale=10)
    path_y.propagate(wf_y)

    obj_aperture.propagate(wf_z)
    obj_aperture.propagate(wf_y)

    wf_z.wavelength /= refr_index
    wf_y.wavelength /= refr_index

    obj_lens.propagate(wf_z)
    obj_lens.propagate(wf_y)

    illumination = np.empty(
        (npix_fov, npix_fov, npix_fov), dtype=wf_z.intensity.dtype)

    # Make sure it is centered even if pixels are odd or even
    offset = 0 if npix_fov % 2 else 0.5

    for pix in range(npix_fov):
        pixel = pix - npix_fov // 2 + offset
        distance = pixel * pixelscale * u.pixel

        psf = poppy.FresnelOpticalSystem()
        psf.add_optic(noop, objective_focal + distance)

        wfc_y = wf_y.copy()
        wfc_z = wf_z.copy()

        psf.propagate(wfc_y)
        psf.propagate(wfc_z)

        resample_wavefront(wfc_y, pixelscale, fov_pixels)
        resample_wavefront(wfc_z, pixelscale, fov_pixels)

        mix = wf_mix(wfc_y, wfc_z)
        mix.normalize()

        illumination[:, pix, :] = mix.intensity

    if rel_thresh is not None:
        illumination = utils.threshold_crop(
            illumination, rel_thresh, 0)

    return illumination / illumination.sum(0).mean()


def gaussian_psf(npix_lateral=129, npix_axial=129,
                 pixelscale=635e-9, wavelength=500e-9,
                 numerical_aperture=0.5, refraction_index=1.33):
    """
    Compute an approximate PSF model based on gaussian beam propagation

    Koskela, O., Montonen, T., Belay, B. et al. Gaussian Light Model in Brightfield
    Optical Projection Tomography. Sci Rep 9, 13934 (2019).
    https://bib-ezproxy.epfl.ch:5295/10.1038/s41598-019-50469-6

    Parameters
    ----------
    npix_lateral : int, optional
        number of pixels in the lateral direction, by default 129
    npix_axial : int, optional
        number of pixels in the axial direction, by default 129
    pixelscale : float, optional
        pixelscale in meters per pixel, by default 1.3e-3/2048
    wavelength : float, optional
        illumination wavelength in meters, by default 500e-9
    numerical_aperture : float, optional
        objective NA, by default 0.5
    refraction_index : float, optional
        imaging medium NA, by default 1.33

    Returns
    -------
    array [ZXY]
        the gaussian PSF

    """

    # compensate for even/odd pixels so that the PSF is always centered
    odd_l = npix_lateral % 2
    odd_a = npix_axial % 2
    lat_offset = 0 if odd_l else 0.5
    ax_offset = 0 if odd_a % 2 else 0.5

    r_coords = (np.arange((npix_lateral + 1) // 2) + lat_offset) * pixelscale
    z_coords = (np.arange((npix_axial + 1) // 2) + ax_offset) * pixelscale

    w0 = wavelength / (np.pi * refraction_index * numerical_aperture)

    z_rayleygh = math.pi * w0 ** 2 * refraction_index / wavelength
    w_z = w0 * np.sqrt(1 + (z_coords/z_rayleygh)**2)
    w_zi2 = 1 / np.square(w_z)
    r_coords = np.square(r_coords)
    intens = np.sqrt(w0**2 * w_zi2)

    gauss_psf = np.einsum('i, ij -> ij', intens,
                          np.exp(- 2 * np.outer(w_zi2, r_coords)))
    gauss_psf = np.einsum('ij, ik->ijk', gauss_psf, gauss_psf)
    gauss_psf = primitives.quadrant_to_volume(gauss_psf, (odd_a, odd_l, odd_l))

    return gauss_psf


if __name__ == '__main__':
    import napari

    s_psf = gaussian_psf(npix_lateral=129, npix_axial=129)
    s_psf = np.log10(s_psf+1e-12)
    illu = openspim_illumination(
        simu_size=1024, npix_fov=256, oversample=8, rel_thresh=1e-6)

    with napari.gui_qt():
        viewer = napari.view_image(s_psf)
        viewer.add_image(illu)



"""
The imaging module provides simulations of different imaging systems
and acquisition techniques for microscopy.

**Conventions:**

arrays follow the ZXY convention, with

    - Z : depth axis (axial, focus axis)
    - X : horizontal axis (lateral)
    - Y : vertical axis (lateral, rotation axis when relevant)

sinograms follow the TPY convention, with

    - T : angles (theta)
    - P : captor axis
    - Y : rotation axis
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import scipy.signal as sig
import numpy as np
import numpy.random as random
from scipy import ndimage
from cbi_toolbox import splineradon as spl


def widefield(obj, psf, pad=False):
    """
    Simulate the widefield imaging of an object

    Parameters
    ----------
    obj : array [ZXY]
        the object to be imaged
    psf : array [ZXY]
        the PSF of the imaging system
    pad : bool, optional
        extend the field of view to see all contributions, by default False
        if False, the image will have the same size as the object

    Returns
    -------
    numpy.ndarray
        the widefield image
    """

    mode = 'full' if pad else 'same'

    image = sig.fftconvolve(obj, psf, mode=mode)
    image.clip(0, None, out=image)
    return image


def opt(obj, psf, theta=None, pad=False):
    """
    Simulate the OPT imaging of an object

    Parameters
    ----------
    obj : array [ZXY]
        the object to be imaged, ZX must be square
    psf : array [ZXY]
        the PSF of the imaging system, Z dimension must match object ZX
    theta : array, optional
        array of rotation angles (in degrees), by default None
        If None, uses numpy.arange(180).
    pad : bool, optional
        extend the field of view to see all contributions
        (needed if the object is not contained in the inner cylinder to the array), by default False

    Returns
    -------
    array [TPY]
        the imaged sinogram

    Raises
    ------
    ValueError
        if the PSF dimension does not match the object
    """

    if not obj.shape[0] == obj.shape[1]:
        raise NotImplementedError(
            'Please provide a square object in ZX dimensions')

    if psf.shape[0] % 2 != obj.shape[0] % 2:
        raise ValueError('In order to correctly center the PSF,'
                         ' please profide a PSF with the same Z axis parity as the object Z axis')

    if theta is None:
        theta = np.arange(180)

    full_size = obj.shape[0]

    if pad:
        full_size = int(full_size * np.sqrt(2)) + 1
        pad_size = (full_size - obj.shape[0]) // 2
        full_size = obj.shape[0] + 2 * pad_size

    if psf.shape[0] < full_size:
        raise ValueError('PSF Z size must be greater than padded object Z size: {} >= {}'.format(
            psf.shape[0], full_size))

    if pad and pad_size:
        obj = np.pad(obj, ((pad_size,), (pad_size,), (0,)))

    if pad:
        mode = 'full'
        sinogram = np.empty(
            (theta.size, obj.shape[0] + psf.shape[1] - 1,
             obj.shape[-1] + psf.shape[-1] - 1))
    else:
        mode = 'same'
        sinogram = np.empty(
            (theta.size, obj.shape[0], obj.shape[-1]))

    splimage = ndimage.spline_filter1d(obj, axis=0)
    splimage = ndimage.spline_filter1d(obj, axis=1, output=splimage)

    crop_size = (psf.shape[0] - obj.shape[0]) // 2
    if crop_size:
        psf = psf[crop_size:-crop_size, ...]

    psf = np.flip(psf, 0)

    for idx, angle in enumerate(theta):
        rotated = ndimage.rotate(
            splimage, angle, prefilter=False, reshape=False)

        rotated = sig.fftconvolve(rotated, psf, axes=(1, 2), mode=mode)

        sinogram[idx, ...] = rotated.sum(0)

    return sinogram


def fps_opt(obj, psf, pad=False, **kwargs):
    """
    Simulate the FPS-OPT (focal plane scanning) imaging of an object

    Parameters
    ----------
    obj : array [ZXY]
        the object to be imaged
    psf : array [ZXY] or array [XY]
        the PSF of the system, or projected PSF along the Z axis
    pad : bool, optional
        pad the vield of view to see all contributions
        (required if the sample is not contained in the inner cylinder of the object), by default False

    **kwargs :
        to be passed to the radon call

    Returns
    -------
    array [TPY]
        the imaged sinogram

    Raises
    ------
    ValueError
        if the PSF is not 2D or 3D
    """

    if psf.ndim == 3:
        psf = psf.sum(0, keepdims=True)
    elif psf.ndim == 2:
        psf = psf[None, ...]
    else:
        raise ValueError("Invalid dimensions for PSF: {}".format(psf.ndim))

    sinogram = spl.radon(obj, circle=(not pad), **kwargs)

    mode = 'full' if pad else 'same'

    image = sig.fftconvolve(sinogram, psf, axes=(1, 2), mode=mode)
    return image


def fss_opt(obj, psf, illu, pad=False, **kwargs):
    """
    Simulate the FSS-OPT (focal sheet scanning) imaging of an object

    Parameters
    ----------
    obj : array [ZXY]
        the object to be imaged
    psf : array [ZXY] or array [XY]
        the PSF of the system
    illu : array [ZXY]
        the illumination function of the SPIM
    pad : bool, optional
        pad the vield of view to see all contributions
        if used, illu will be required to be bigger
        (required if the sample is not contained in the inner cylinder of the object), by default False

    **kwargs :
        to be passed to the radon call

    Returns
    -------
    array [TPY]
        the imaged sinogram

    Raises
    ------
    ValueError
        if the PSF dimensions do not fit the illumination
        if the illumination function has an incorrect shape
    """

    if pad:
        # TODO use bigger SPIM illumination function to fit padded sinogram
        raise NotImplementedError()

    if psf.shape[0] % 2 != illu.shape[0] % 2:
        raise ValueError('In order to correctly center the illumination on the PSF,'
                         ' please profide a PSF with the same Z axis parity as the illumination Z axis')

    if not (psf.shape[1] % 2 and psf.shape[2] % 2):
        raise ValueError('In order to correctly center the PSF on the pixels,'
                         ' please provide a PSF with odd parity on X and Y axis')

    crop = (psf.shape[0] - illu.shape[0]) // 2
    if crop:
        psf = psf[crop:-crop, ...]
    psf = np.flip(psf)

    sinogram = spl.radon(obj, circle=(not pad), **kwargs)

    psf_xw = psf.shape[1]
    psf_yw = psf.shape[2]
    pad_x = psf_xw // 2
    pad_y = psf_yw // 2

    illu = np.pad(illu, ((0,), (pad_x,), (pad_y,)))

    image = np.zeros_like(sinogram, shape=(
        sinogram.shape[0], sinogram.shape[1] + psf_xw - 1, sinogram.shape[2] + psf_yw - 1))

    for x in range(sinogram.shape[1]):
        for y in range(sinogram.shape[2]):
            local = np.sum(illu[:, x:x+psf_xw, y:y+psf_yw] * psf, 0)
            pixel = sinogram[:, x, y]
            spread = pixel[:, None, None] * local[None, ...]

            image[:, x:x+psf_xw, y:y+psf_xw] += spread

    if not pad:
        image = image[:, pad_x:-pad_x, pad_y:-pad_y]

    return image


def spim(obj, psf, illu, pad=False):
    """
    Simulate the SPIM imaging of an object

    Parameters
    ----------
    obj : array [ZXY]
        the object to be imaged
    psf : array [ZXY]
        the PSF of the imaging objective
    illu : array [ZXY]
        the illumination function of the SPIM
    pad : bool, optional
        pad the vield of view to see all contributions, by default False
    """

    if psf.shape[0] % 2 != illu.shape[0] % 2:
        raise ValueError('In order to correctly center the illumination on the PSF,'
                         ' please profide a PSF with the same Z axis parity as the illumination Z axis')

    if psf.shape[0] < illu.shape[0]:
        raise ValueError(
            'PSF depth must be bigger than SPIM illumination depth (Z axis) - {} < {}'.format(psf.shape[0], illu.shape[0]))

    if illu.shape[1] != obj.shape[1] or illu.shape[2] != obj.shape[2]:
        raise ValueError(
            'SPIM illumination XY must match object XY dimensions - [{},{}] != [{},{}]'.format(
                illu.shape[1], illu.shape[2], obj.shape[1], obj.shape[2]))

    thickness = illu.shape[0]
    crop = (psf.shape[0] - thickness) // 2
    if crop:
        psf = psf[crop:-crop, ...]
    psf = np.flip(psf, 0)

    if pad:
        image = np.empty(
            (obj.shape[0] + thickness - 1, obj.shape[1] + psf.shape[1] - 1, obj.shape[2] + psf.shape[2] - 1))
        mode = 'full'
        obj = np.pad(obj, ((thickness - 1,), (0,), (0,)))

    else:
        image = np.empty(
            (obj.shape[0], obj.shape[1], obj.shape[2]))
        mode = 'same'
        obj = np.pad(obj, ((thickness // 2,), (0,), (0,)))

    for index in range(image.shape[0]):
        sliced = obj[index:index+thickness, ...] * illu
        sliced = sig.fftconvolve(sliced, psf, axes=(1, 2), mode=mode)
        image[index, ...] = sliced.sum(0)

    return image


def noise(image, photons=200, background=3, seed=None, in_place=False):
    """
    Simulate the acquisition noise of the camera:
        noise = shot noise

    Parameters
    ----------
    image : array [ZXY]
        the clean image
    photons : int, optional
        the max level of photons per pixel, by default 150
    background : int, optional
        the background level of photons per pixel, by default 3
    seed : int, optional
        the seed for rng, by default None

    Returns
    -------
    array [ZXY]
        the noisy image
    """

    if not in_place:
        image = image.copy()

    amp = image.max()

    image *= (photons - background) / amp
    image += background
    rng = random.default_rng(seed)
    poisson = rng.poisson(image)

    return amp * poisson / photons


if __name__ == "__main__":
    from cbi_toolbox.simu import primitives, optics
    import napari
    import time

    TEST_SIZE = 64

    sample = primitives.boccia(
        TEST_SIZE, radius=(0.8 * TEST_SIZE) // 2, n_stripes=4)
    s_psf = optics.gaussian_psf(
        numerical_aperture=0.3,
        npix_axial=TEST_SIZE+1, npix_lateral=TEST_SIZE+1)
    opt_psf = optics.gaussian_psf(
        numerical_aperture=0.1, npix_axial=TEST_SIZE, npix_lateral=TEST_SIZE+1)
    spim_illu = optics.openspim_illumination(
        slit_opening=2e-3,
        npix_fov=TEST_SIZE, simu_size=4*TEST_SIZE, rel_thresh=1e-3)
    s_theta = np.arange(90)

    start = time.time()
    s_widefield = widefield(sample, s_psf)
    print('Time for widefield: \t{}s'.format(time.time() - start))
    start = time.time()
    noisy = noise(s_widefield)
    print('Time for noise: \t{}s'.format(time.time() - start))

    start = time.time()
    s_spim = spim(sample, s_psf, spim_illu)
    print('Time for SPIM: \t{}s'.format(time.time() - start))

    start = time.time()
    s_opt = opt(sample, opt_psf, theta=s_theta)
    print('Time for OPT: \t{}s'.format(time.time() - start))

    start = time.time()
    s_fpsopt = fps_opt(sample, s_psf, theta=s_theta)
    print('Time for FPS-OPT: \t{}s'.format(time.time() - start))

    start = time.time()
    s_fssopt = fss_opt(sample, s_psf, spim_illu, theta=s_theta)
    print('Time for FSS-OPT: \t{}s'.format(time.time() - start))

    start = time.time()
    s_radon = spl.radon(sample, theta=s_theta, circle=True)
    print('Time for radon: \t{}s'.format(time.time() - start))

    with napari.gui_qt():
        viewer = napari.view_image(sample)
        viewer.add_image(s_widefield)
        viewer.add_image(noisy)
        viewer.add_image(s_spim)

        viewer = napari.view_image(s_radon)
        viewer.add_image(s_opt)
        viewer.add_image(s_fpsopt)
        viewer.add_image(s_fssopt)



# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause



"""
This module implements change of basis for bsplines.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>,
# Christian Jaques <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is a translation of Michael Liebling's matlab code,
# which was already largely based on a C-library written by Philippe
# Thevenaz, BIG, EPFL

import numpy as np

from ._interpolation_conversion import convert_to_samples, convert_to_interpolation_coefficients
from cbi_toolbox.utils import transpose_dim_to


def change_basis(in_array, from_basis, to_basis, degree, axes=(0,),
                 tolerance=1e-9, boundary_condition='Mirror',
                 in_place=False):
    """
    Change the basis along the provided axes
    (e.g. to convert a 2D signal, use axes=(0, 1)).

    Parameters
    ----------
    in_array : numpy.ndarray
        The input signal.
    from_basis : str
        Source basis, one of ['cardinal', 'dual', b-spline'] (case insensitive).
    to_basis : str
        Target basis, one of ['cardinal', 'dual', b-spline'] (case insensitive).
    degree : int
        Degree of the basis.
    axes : tuple, optional
        Axes on which to compute the change of basis, by default (0,).
    tolerance : float, optional
        Precision of the computations, by default 1e-9.
    boundary_condition : str, optional
        Boundary conditions, one of ['mirror', 'periodic'], by default 'Mirror'.
        (case insensitive)
    in_place : bool, optional
        Overwrite the source data to save memory space, by default False.

    Returns
    -------
    numpy.ndarray
        The signal in the target base.
    """

    if not in_place:
        in_array = in_array.copy()

    try:
        for ax in axes:
            in_array = change_basis(
                in_array, from_basis, to_basis, degree, ax, tolerance, boundary_condition)
    except TypeError:
        in_array = transpose_dim_to(in_array, axes, 0)
        in_array = change_basis_inner(
            in_array, from_basis, to_basis, degree, tolerance, boundary_condition)
        in_array = transpose_dim_to(in_array, 0, axes)

    return in_array


def change_basis_inner(in_array, from_basis, to_basis, degree, tolerance=1e-9,
                       boundary_condition='Mirror'):
    """
    Inner implementation of the change the basis.
    Applies only on the first axis, and always in-place.

    Parameters
    ----------
    in_array : numpy.ndarray
        The input signal.
    from_basis : str
        Source basis, one of ['cardinal', 'dual', b-spline'] (case insensitive).
    to_basis : str
        Target basis, one of ['cardinal', 'dual', b-spline'] (case insensitive).
    degree : int
        Degree of the basis.
    tolerance : float, optional
        Precision of the computations, by default 1e-9.
    boundary_condition : str, optional
        Boundary conditions, one of ['mirror', 'periodic'], by default 'Mirror'.
        (case insensitive)

    Returns
    -------
    numpy.ndarray
        The signal in the target base.
    """

    from_basis = from_basis.upper()
    to_basis = to_basis.upper()

    if from_basis == 'CARDINAL':
        if to_basis == 'CARDINAL':
            return in_array
        elif to_basis == 'B-SPLINE':
            output = convert_to_interpolation_coefficients(in_array, degree, tolerance,
                                                           boundary_condition=boundary_condition)
        elif to_basis == 'DUAL':
            output = change_basis_inner(in_array, 'cardinal', 'b-spline', degree, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
            output = change_basis_inner(output, 'b-spline', 'dual', degree, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
        else:
            raise ValueError("Illegal to_basis : {0}".format(to_basis))

    elif from_basis == 'B-SPLINE':
        if to_basis == 'CARDINAL':
            output = convert_to_samples(in_array, degree,
                                        boundary_condition=boundary_condition)
        elif to_basis == 'B-SPLINE':
            return in_array
        elif to_basis == 'DUAL':
            output = change_basis_inner(in_array, 'b-spline', 'cardinal', 2 * degree + 1, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
        else:
            raise ValueError("Illegal to_basis : {0}".format(to_basis))

    elif from_basis == 'DUAL':
        if to_basis == 'CARDINAL':
            output = change_basis_inner(in_array, 'dual', 'b-spline', degree, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
            output = change_basis_inner(output, 'b-spline', 'cardinal', degree, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
        elif to_basis == 'B-SPLINE':
            output = change_basis_inner(in_array, 'cardinal', 'b-spline', 2 * degree + 1, tolerance=tolerance,
                                        boundary_condition=boundary_condition)
        elif to_basis == 'DUAL':
            return in_array
        else:
            raise ValueError("Illegal to_basis : {0}".format(to_basis))

    else:
        raise ValueError("Illegal from_basis : {0}".format(from_basis))

    return output


if __name__ == "__main__":

    data = np.random.default_rng().random((100, 100))
    toleranc = 1e-12
    degre = 3
    axe = (0, 1)
    condition = 'periodic'

    c_out = change_basis(data, 'cardinal', 'dual', degre, axes=axe, tolerance=toleranc, boundary_condition=condition,
                         in_place=False)
    ar = change_basis(c_out, 'dual', 'cardinal', degre, axes=axe,
                      tolerance=toleranc, boundary_condition=condition)

    print('Relative error is ', np.max(
        np.linalg.norm((data - ar) / np.abs(data))))
    print('are samples back to signal? ', np.allclose(data, ar))



# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause


import os
import unittest
import numpy as np
from cbi_toolbox import utils
try:
    from cbi_toolbox.parallel import mpi
    MPI_AVAILABLE = True
except ImportError:
    mpi = None
    MPI_AVAILABLE = False


def check_distributed_array(reference, array, axis):
    """
    Check that an array is correctly distributed on given axis

    :param reference: the full array
    :param array: the distributed array
    :param axis:
    :return: distribution is correct (bool)
    """

    s_index, b_size = mpi.distribute_mpi(reference.shape[axis])
    check_mat = utils.transpose_dim_to(reference, axis, 0)
    check_mat = check_mat[s_index:s_index + b_size]
    check_mat = utils.transpose_dim_to(check_mat, 0, axis)
    return np.all(check_mat == array)


def test_distributed_save(file_name, reference, array, axis):
    mpi.save(file_name, array, axis, reference.shape)
    if mpi.is_root_process():
        reloaded_array = np.load(file_name)
        return np.all(reference == reloaded_array)
    else:
        return True


@unittest.skipUnless(MPI_AVAILABLE, "MPI optional dependency not installed.")
class TestDistributedArrays(unittest.TestCase):
    ref_file = 'test_source.npy'
    tmp_file = 'test_save.npy'
    reference = None
    dims = [2, 3, 4, 5]
    ndims = None

    @classmethod
    def setUpClass(cls):
        cls.ndims = len(cls.dims)
        cls.reference = np.empty(cls.dims, dtype=np.float64)
        cls.reference.flat = np.arange(
            int(np.prod(cls.dims)), dtype=np.float64)

        if mpi.is_root_process():
            np.save(cls.ref_file, cls.reference)
        mpi.wait_all()

    @classmethod
    def tearDownClass(cls):
        if mpi.is_root_process():
            os.remove(cls.ref_file)

    def test_load_array(self):
        for axis in range(TestDistributedArrays.ndims):
            sl_array, f_shape = mpi.load(
                TestDistributedArrays.ref_file, axis)
            self.assertTrue(np.array_equal(
                f_shape, TestDistributedArrays.dims))
            self.assertTrue(check_distributed_array(
                TestDistributedArrays.reference, sl_array, axis))

    def test_distribute_array(self):
        for src_axis in range(TestDistributedArrays.ndims):
            sl_array, f_shape = mpi.load(
                TestDistributedArrays.ref_file, src_axis)

            for tgt_axis in range(TestDistributedArrays.ndims):
                temp_array = mpi.redistribute(
                    sl_array, src_axis, tgt_axis, f_shape)
                self.assertTrue(check_distributed_array(
                    TestDistributedArrays.reference, temp_array, tgt_axis))

    def test_save_array(self):
        for axis in range(TestDistributedArrays.ndims):
            sl_array, f_shape = mpi.load(
                TestDistributedArrays.ref_file, axis)
            mpi.save(TestDistributedArrays.tmp_file,
                     sl_array, axis, f_shape)

            if mpi.is_root_process():
                reloaded_array = np.load(TestDistributedArrays.tmp_file)
                self.assertTrue(np.array_equal(
                    TestDistributedArrays.reference, reloaded_array))
                os.remove(TestDistributedArrays.tmp_file)


if __name__ == "__main__":
    unittest.main()



"""
This module implements the computation of a bspline kernel for FBP interpolation.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
# and Michael Liebling <michael.liebling@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is inspired from `SPLRADON` written by Michael Liebling
# <http://sybil.ece.ucsb.edu/pages/splineradon/splineradon.html>


import numpy as np
import scipy


def get_kernel_table(nt, n1, n2, h, s, angles, degree=True):
    """
    Compute the bspline kernel table used for interpolation in FBP.

    Parameters
    ----------
    nt : int
        Number of points in the kernel.
    n1 : int
        Degree of the first spline basis.
    n2 : int
        Degree of the second spline basis.
    h : float
        Pixel sampling step on the first signal.
    s : float
        Pixel sampling step on the second signal.
    angles : array_like
        Angle at which to compute the kernel.
    degree : bool, optional
        Give angles in degrees instead of radians, by default True.

    Returns
    -------
    (numpy.ndarray, float)
        The kernel lookup table and its maximum element.
    """

    pad_fact = 4
    angles = np.atleast_1d(angles)

    if degree:
        angles = np.deg2rad(angles)

    h1 = np.abs(np.sin(angles) * h)
    h2 = np.abs(np.cos(angles) * h)

    a = np.max(h1 * (n1 + 1) / 2 + h2 * (n1 + 1) / 2 + s * (n2 + 1) / 2)

    n2, n3 = n1, n2
    h3 = s

    T = a / (nt - 1)
    dnu = 1 / (T * (pad_fact * nt - 1))
    nu = -1 / (2 * T) + np.arange(pad_fact * nt) * dnu

    trikernel_hat = np.power(np.sinc(np.outer(h1, nu)), (n1 + 1)) * np.power(
        np.sinc(np.outer(h2, nu)), (n2 + 1)) * np.power(np.sinc(np.outer(h3, nu)), (n3 + 1))

    kernel = np.abs(scipy.fft.rfft(trikernel_hat, axis=1, overwrite_x=True))

    table = kernel[:, 0:nt] / (T * nt * pad_fact)

    return table, a



"""
This modules implements signal conversion to bspline interpolation coefficients
and back.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>,
# Christian Jaques <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is a translation of Michael Liebling's matlab code,
# which was already largely based on a C-library written by Philippe
# Thevenaz, BIG, EPFL

import math

import numpy as np
from scipy import signal
from scipy import ndimage

from cbi_toolbox.utils import make_broadcastable
from cbi_toolbox import bsplines
from cbi_toolbox import parallel


def initial_causal_coefficient(coeff, z, tolerance, boundary_condition='mirror'):
    """
    Computes the initial causal coefficient from an array of coefficients
    In the input array, signals are considered to be along the first dimension
    (1D computations).
    """

    n = coeff.shape[0]

    # mirror boundaries condition (mirror is on last sample)
    horizon = n
    if tolerance > 0.0:
        horizon = int(math.ceil(math.log(tolerance) / math.log(abs(z))))

    if boundary_condition.upper() == 'MIRROR':
        if horizon >= n:
            horizon = n
            # vectorization of exponentials of z
            z_powers = np.arange(horizon - 2) + 1
            z_exp = np.ones(horizon - 2) * z
            z_exp = np.power(z_exp, z_powers)
            # the whole signal is taken into account, boundary conditions
            # have to be taken into account. Mirror conditions -->
            z_exp_mirror = np.ones(horizon - 2) * z
            z_powers = np.arange(2 * n - 3, n - 1, -1)
            z_exp_mirror = np.power(z_exp_mirror, z_powers)
            z_exp += z_exp_mirror

            z_exp = make_broadcastable(z_exp, coeff)

            # compute sum c[k]*z**k
            c = np.sum(coeff[1:horizon - 1, ...] * z_exp, axis=0)
            c /= (1 - z ** (2 * n - 1))
            c += coeff[0, ...] + coeff[-1, ...] * z ** (n - 1)
        else:
            # vectorization of exponentials of z
            z_powers = np.arange(horizon)
            z_exp = np.ones(horizon) * z
            z_exp = np.power(z_exp, z_powers)

            z_exp = make_broadcastable(z_exp, coeff)

            # compute sum c[k]*z**k
            c = np.sum(coeff[:horizon, ...] * z_exp, axis=0)

    elif boundary_condition.upper() == 'PERIODIC':
        temp_shape = list(coeff.shape)
        temp_shape[0] = 1
        temp = np.ones(temp_shape)
        temp[0, ...] = coeff[0, ...]
        c1 = np.concatenate((temp, coeff[-1:0:-1, ...]), axis=0)
        if tolerance > 0.0:
            horizon2 = np.mod(np.arange(horizon), n)
            # vectorization of exponentials of z
            z_powers = np.arange(horizon)
            z_exp = np.ones(horizon) * z
            z_exp = np.power(z_exp, z_powers)
            # the whole signal is taken into account, boundary conditions
            # have to be taken into account. Periodic conditions -->
            z_exp = make_broadcastable(z_exp, c1)

            # compute sum c[k]*z**k
            c = np.sum(c1[horizon2, ...] * z_exp, axis=0)
        else:
            # vectorization of exponentials of z
            z_powers = np.arange(horizon - 1) + 1
            z_exp = np.ones(horizon - 1) * z
            z_exp = np.power(z_exp, z_powers)
            z_exp = make_broadcastable(z_exp, c1)

            # compute sum c[k]*z**k
            c = c1[0, ...] + \
                np.sum(np.multiply(c1[1:, ...], z_exp), axis=0) / (1 - z ** n)

    else:
        raise ValueError('Illegal boundary condition: {}'.format(
            boundary_condition.upper()))

    return c


def initial_anticausal_coefficient(c, z, boundary_condition='Mirror'):
    """
    Computes the initial anti-causal coefficient from an array of coefficients
    In the input array, signals are considered to be along the first dimension
    (1D computations).
    """
    if boundary_condition.upper() == 'MIRROR':
        c0 = -z * (c[-1, ...] + z * c[-2, ...]) / (1 - z ** 2)
    elif boundary_condition.upper() == 'PERIODIC':
        n = c.shape[0]
        z_exp = np.ones(n) * z
        z_power = np.arange(n, 0, -1)
        z_exp = np.power(z_exp, z_power)

        temp_shape = list(c.shape)
        temp_shape[0] = 1
        temp = np.ones(temp_shape)

        temp[0] = c[-1, ...]
        c1 = np.concatenate((c[-2::-1, ...], temp), axis=0)

        z_exp = make_broadcastable(z_exp, c1)

        c0 = (1. / (z ** n - 1.)) * np.sum(np.multiply(c1, z_exp), axis=0)
    else:
        raise ValueError('Illegal boundary condition: {}'.format(
            boundary_condition.upper()))
    return c0


def convert_to_interpolation_coefficients(c, degree, tolerance=1e-9,
                                          boundary_condition='Mirror',
                                          in_place=False):
    """
    Computes the b-spline interpolation coefficients of a signal
    In the input array, the signals are considered along the first
    dimension (1D computations).
    """

    if degree == 0 or degree == 1 or c.shape[0] == 1:
        if not in_place:
            return c.copy()
        return c

    elif 2 <= degree <= 5:
        if boundary_condition.upper() == 'MIRROR':
            mode = 'mirror'
        elif boundary_condition.upper() == 'PERIODIC':
            mode = 'grid-wrap'
        else:
            raise ValueError(
                'Invalid boundary condition: {}'.format(boundary_condition))

        output = None
        if in_place:
            output = c

        if c.ndim < 2:
            output = ndimage.spline_filter1d(
                c, degree, axis=0, mode=mode, output=output)

        else:
            if output is None:
                output = np.empty_like(c)

            split_size = c.shape[-1]

            def to_spline(start, width):
                in_array = c[:, ..., start:start+width]
                out_array = output[:, ..., start:start+width]
                ndimage.spline_filter1d(
                    in_array, degree, axis=0, mode=mode, output=out_array)

            parallel.parallelize(to_spline, split_size)

        return output

    if not in_place:
        c = c.copy()

    if degree == 6:
        z = [-0.488294589303044755130118038883789062112279161239377608394,
             -0.081679271076237512597937765737059080653379610398148178525368,
             -0.00141415180832581775108724397655859252786416905534669851652709]
    elif degree == 7:
        z = [-0.5352804307964381655424037816816460718339231523426924148812,
             -0.122554615192326690515272264359357343605486549427295558490763,
             -0.0091486948096082769285930216516478534156925639545994482648003]
    elif degree == 8:
        z = [-0.57468690924876543053013930412874542429066157804125211200188,
             -0.163035269297280935240551896860737052234768145508298548489731,
             -0.0236322946948448500234039192963613206126659208546294370651457,
             -0.000153821310641690911739352530184021607629640540700430019629940]
    elif degree == 9:
        z = [-0.60799738916862577900772082395428976943963471853990829550220,
             -0.201750520193153238796064685055970434680898865757470649318867,
             -0.043222608540481752133321142979429688265852380231497069381435,
             -0.00212130690318081842030489655784862342205485609886239466341517]
    elif degree == 10:
        z = [-0.63655066396942385875799205491349773313787959010128860432339,
             -0.238182798377573284887456162200161978666543494059728787251924,
             -0.065727033228308551538201803949684252205121694392255863103034,
             -0.0075281946755486906437698340318148831650567567441314086093636,
             -0.0000169827628232746642307274679399688786114400132341362095006930]
    elif degree == 11:
        z = [-0.66126606890073470691013126292248166961816286716800880802421,
             -0.272180349294785885686295280258287768151235259565335176244192,
             -0.089759599793713309944142676556141542547561966017018544406214,
             -0.0166696273662346560965858360898150837154727205519335156053610]
    else:
        raise ValueError("Invalid spline degree {0}".format(degree))

    # compute overall gain
    z = np.atleast_1d(z)
    # apply gain to coeffs
    c *= np.prod((1 - z) * (1 - 1 / z))

    # loop over all poles
    for pole in z:
        # causal initialization
        c[0, ...] = initial_causal_coefficient(
            c, pole, tolerance, boundary_condition)
        # causal filter
        zinit = pole * c[0, ...]
        zinit = zinit[np.newaxis, ...]

        if c.ndim < 2:
            c[1:, ...], _ = signal.lfilter(
                [1], [1, -pole], c[1:, ...], axis=0, zi=zinit)

        else:
            split_size = c.shape[-1]

            def lfilt(start, width):
                c[1:, ..., start:start+width], _ = signal.lfilter(
                    [1], [1, -pole], c[1:, ..., start:start+width], axis=0, zi=zinit[:, ..., start:start+width])

            parallel.parallelize(lfilt, split_size)

        # anticausal initialization
        c[-1, ...] = initial_anticausal_coefficient(
            c, pole, boundary_condition=boundary_condition)
        # anticausal filter
        zinit = pole * c[-1, ...]
        zinit = zinit[np.newaxis, ...]

        if c.ndim < 2:
            c[:-1], _ = signal.lfilter([-pole], [1, -pole],
                                       np.flipud(c[0:-1, ...]), axis=0, zi=zinit)

        else:
            split_size = c.shape[-1]
            fc = np.flipud(c[0:-1, ...])

            def lfilt2(start, width):
                c[:-1, ...,  start:start+width], _ = signal.lfilter([-pole], [1, -pole],
                                                                    fc[:, ...,  start:start+width], axis=0, zi=zinit[:, ..., start:start+width])

            parallel.parallelize(lfilt2, split_size)

        c[:-1] = np.flipud(c[:-1])

    return c


def convert_to_samples(c, deg, boundary_condition='Mirror', in_place=False):
    """
    Convert interpolation coefficients into samples
    In the input array, the signals are considered along the first dimension
    (1D computations).
    """

    n = c.shape[0]
    if n == 1:
        return c

    kerlen = int(2 * math.floor(deg / 2.) + 1)

    k = -math.floor(deg / 2.) + np.arange(kerlen)
    kernel = bsplines.bspline(k, deg)

    # different extensions based on boundary condition
    if boundary_condition.upper() == 'MIRROR':
        boundary = 'mirror'
    elif boundary_condition.upper() == 'PERIODIC':
        boundary = 'wrap'
    else:
        raise ValueError('Illegal boundary condition: {}'.format(
            boundary_condition.upper()))

    kernel = make_broadcastable(kernel, c)

    output = None
    if in_place:
        output = c

    return ndimage.convolve(c, kernel, mode=boundary, output=output)



"""
The plotting module contains helper functions to plot animated movies in matplotlib.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import matplotlib.pyplot as plt
import matplotlib.animation as animation


class AnimImshow:
    """
    Create an animated figure over an array of images.

    Parameters
    ----------
    images : np.ndarray
        Array containing the images, shape [t, x, y].
    interval : int, optional
        Frame duration in ms, by default 100.
    """

    def __init__(self, images, interval=100):
        self.interval = interval
        self.images = images
        self.fig = plt.figure()
        self.im = plt.imshow(images[0, ...], animated=True)
        self.ani = animation.FuncAnimation(
            self.fig, self._updatefig, interval=interval, frames=images.shape[0], blit=True)

    def _updatefig(self, anim_index):
        self.im.set_array(self.images[anim_index, ...])
        return self.im

    def save_to_gif(self, path):
        """
        Save the animated figure as a gif file.

        Parameters
        ----------
        path : str
            Path to the file to save.
        """

        self.ani.save(path, writer='imagemagick', fps=1000/self.interval)



"""
The bspline package implements splines and transformations to project signals
onto spline bases based on formulas described in [1].

[1] P. Thévenaz, T. Blu, M. Unser, *"Interpolation Revisited"*, IEEE Transactions
on Medical Imaging, vol. 19, no. 7, pp. 739-758, July 2000.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>,
# Christian Jaques <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is a translation of Michael Liebling's matlab code,
# which was already largely based on a C-library written by Philippe
# Thevenaz, BIG, EPFL

import numpy as np
from numpy import zeros, power, multiply
from scipy import signal

from ._change_basis import *
from ._interpolation_conversion import *


def bspline(x, deg):
    """
    B-spline function, faster for degree up to 7.

    Parameters
    ----------
    x : numpy.ndarray
        The points at which the bspline is evaluated.
    deg : int
        The degree of the bspline.

    Returns
    -------
    numpy.ndarray
        The values of the spline for each x.
    """

    if deg == 0:
        return _bspline00(x)
    elif deg == 1:
        return _bspline01(x)
    elif deg == 2:
        return _bspline02(x)
    elif deg == 3:
        return _bspline03(x)
    elif deg == 4:
        return _bspline04(x)
    elif deg == 5:
        return _bspline05(x)
    elif deg == 6:
        return _bspline06(x)
    elif deg == 7:
        return _bspline07(x)
    else:
        return signal.bspline(x, deg)


def _bspline00(x):
    """
    Bspline of degree 0.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    idx = np.abs(x) == 0.5
    y[idx] = 0.5
    idx = np.abs(x) < 0.5
    y[idx] = 1.

    return y


def _bspline01(x):
    """
    Bspline of degree 1.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    idx = np.abs(x) < 1
    y[idx] = 1 - abs(x[idx])

    return y


def _bspline02(x):
    """
    Bspline of degree 2.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    idx = np.abs(x) < 0.5
    y[idx] = 0.75 - power(x[idx], 2)
    idx = (np.abs(x) < 1.5) & (np.abs(x) >= 0.5)
    y[idx] = 0.5 * power((np.abs(x[idx]) - 1.5), 2)

    return y


def _bspline03(x):
    """
    Bspline of degree 3.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    idx = np.abs(x) < 1.
    y[idx] = 0.5 * (power(np.abs(x[idx]), 3) - 2. * power(x[idx], 2)) + 2. / 3.
    idx = (np.abs(x) < 2.) & (np.abs(x) >= 1.)
    y[idx] = -1. / 6. * (power(np.abs(x[idx]) - 2, 3))

    return y


def _bspline04(x):
    """
    Bspline of degree 4.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    a = power(x, 2)
    idx = np.abs(x) < 0.5
    y[idx] = (multiply(a[idx], a[idx] * 0.25 - 5. / 8.) + 115. / 192)
    idx = (np.abs(x) < 1.5) & (np.abs(x) >= 0.5)
    y[idx] = (multiply(np.abs(x[idx]), multiply(np.abs(x[idx]), multiply(np.abs(x[idx]), 5. / 6. - np.abs(
        x[idx]) * 1. / 6.) - 5. / 4.) + 5. / 24.) + 55. / 96.)
    idx = (np.abs(x) < 2.5) & (np.abs(x) >= 1.5)
    a = power(np.abs(x) - 5. / 2., 4)
    y[idx] = (a[idx] * 1. / 24.)

    return y


def _bspline05(x):
    """
    Bspline of degree 5.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    a = power(x, 2)
    idx = np.abs(x) < 1.
    y[idx] = (multiply(a[idx], multiply(a[idx], 1. / 4. -
                                        np.abs(x[idx]) * 1. / 12.) - 1. / 2.) + 11. / 20.0)
    idx = (np.abs(x) < 2.) & (np.abs(x) >= 1.)
    a = np.abs(x)
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], a[
        idx] * 1.0 / 24.0 - 3. / 8.) + 5. / 4.) - 7. / 4.) + 5. / 8.) + 17. / 40.)
    idx = (np.abs(x) < 3.) & (np.abs(x) >= 2.)
    a = 3. - np.abs(x)
    y[idx] = (power(a[idx], 5) * 1. / 120.)

    return y


def _bspline06(x):
    """
    Bspline of degree 6.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    idx = np.abs(x) < 0.5
    a = power(x, 2)
    y[idx] = (multiply(a[idx], multiply(
        a[idx], (7. / 48. - a[idx] * 1. / 36)) - 77. / 192.) + 5887. / 11520.)
    idx = (np.abs(x) < 1.5) & (np.abs(x) >= 0.5)
    a = np.abs(x)
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], a[
        idx] * 1. / 48. - 7. / 48.) + 21. / 64.) - 35. / 288.) - 91. / 256) - 7. / 768.) + 7861. / 15360.)
    idx = (np.abs(x) < 2.5) & (np.abs(x) >= 1.5)
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], 7. / 60. - a[
        idx] * 1. / 120.) - 21. / 32.) + 133. / 72.) - 329. / 128.0) + 1267. / 960.) + 1379. / 7680.)
    idx = (np.abs(x) < 3.5) & (np.abs(x) >= 2.5)
    y[idx] = (power(a[idx] - 7. / 2., 6) * 1. / 720.)

    return y


def _bspline07(x):
    """
    Bspline of degree 7.

    Parameters
    ----------
    x : numpy.ndarray
        The points at whichthe bspline is evaluated.

    Returns
    -------
    numpy.ndarray
        The values of the bspline for each x.
    """

    y = zeros(x.shape)
    a = power(x, 2)
    idx = np.abs(x) < 1.
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], np.abs(
        x[idx]) * 1. / 144. - 1. / 36.) + 1. / 9.) - 1. / 3.) + 151. / 315.)
    idx = (np.abs(x) < 2.) & (np.abs(x) >= 1.)
    a = np.abs(x)
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx],
                                                                                                            1. / 20. -
                                                                                                            a[idx] * 1. / 240.) - 7. / 30.) + 1. / 2.) - 7. / 18.) - 1. / 10.) - 7. / 90.) + 103. / 210.)
    idx = (np.abs(x) < 3.) & (np.abs(x) >= 2.)
    y[idx] = (multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], multiply(a[idx], a[
        idx] * 1. / 720. - 1. / 36.) + 7. / 30.) - 19. / 18.) + 49. / 18.) - 23. / 6.) + 217. / 90.) - 139. / 630.)
    idx = (np.abs(x) < 4.) & (np.abs(x) >= 3)
    y[idx] = power(4. - a[idx], 7) * 1. / 5040.

    return y



"""
The texture module allows to generate 3D textures for synthetic samples
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import noise
from cbi_toolbox.simu import primitives


def spheres(size, density=1, seed=None):
    """
    Generates a texture full of hollow spheres

    Parameters
    ----------
    size : int
        size of the texture
    density : int, optional
        spheres density in the texture, by default
    seed : int, optional
        seed of the rng, default is None

    Returns
    -------
    array [size, size, size]
        the texture
    """

    dtype = np.float32
    n_spheres = int(density * 10000)

    max_radius = int(0.1 * size)
    min_radius = int(0.02 * size)

    max_in_radius = 0.5

    min_intens = 0.05
    max_intens = 0.2

    pad_size = size + 4 * max_radius

    volume = np.ones((pad_size, pad_size, pad_size), dtype=dtype)

    rng = np.random.default_rng(seed)

    for _ in range(n_spheres):
        center = (rng.random(3) * (size + 2 * max_radius)
                  ).astype(int) + max_radius

        radius = int(rng.uniform(min_radius, max_radius))
        in_radius = rng.uniform(0, max_in_radius)
        intens = rng.uniform(min_intens, max_intens)

        obj = primitives.ball(radius * 2, in_radius=in_radius, dtype=dtype)

        volume[center[0] - radius:center[0] + radius, center[1] - radius:center[1] + radius,
               center[2] - radius:center[2] + radius] *= (1 - obj * intens)

    return 1 - volume[volume.ndim * [slice(2*max_radius, -2*max_radius)]]


def simplex(size, scale=1, octaves=3, persistence=0.7, lacunarity=3.5, seed=None):
    """
    Generates 3D simplex noise

    Parameters
    ----------
    size : int
        size of the texture
    scale : int, optional
        scale of the noise, by default 1
    octaves : int, optional
        number of octaves used, by default 3
    persistence : float, optional
        relative amplitude of octaves, by default 0.7
    lacunarity : float, optional
        relative frequency of octaves, by default 3.5
    seed : int, optional
        seed for the noise, by default None

    Returns
    -------
    array [size, size, size]
        the texture
    """

    if seed is None:
        seed = int(np.random.default_rng().integers(2**10) * scale)

    volume = np.empty((size, size, size), dtype=np.float32)
    scale /= size

    # TODO optimize loops
    for x in range(size):
        for y in range(size):
            for z in range(size):
                sample = noise.snoise3(seed + x*scale, seed + y*scale, seed + z*scale,
                                       octaves=octaves, persistence=persistence, lacunarity=lacunarity)
                volume[x, y, z] = sample
    return volume


if __name__ == '__main__':
    import napari

    TEST_SIZE = 128
    s_spheres = spheres(TEST_SIZE)
    s_simplex = simplex(TEST_SIZE)

    phantom = primitives.phantom(TEST_SIZE) * (s_simplex * 0.5 + 0.75)

    with napari.gui_qt():
        viewer = napari.view_image(s_spheres)
        viewer.add_image(s_simplex)
        viewer.add_image(phantom)



"""
This module contains utility functions to work on numpy arrays.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np


def roll_array(array, shifts, axis):
    """
    Roll an array over a given axis, with a different shift for each element
    along the first axis.
    This applies numpy.roll to each sub-array with the corresponding shift.

    Parameters
    ----------
    array : numpy.ndarray
        The original array.
    shifts : list(int)
        The shifts to apply, should be the same size as array.shape[0].
    axis : int
        The axis on which to apply the roll (>0).

    Returns
    -------
    numpy.ndarray
        The rolled array.
    """

    axis -= 1
    rolled = np.empty_like(array)

    for index, array_2D in enumerate(array):
        rolled[index, ...] = np.roll(array_2D, shifts[index], axis=axis)

    return rolled


def center_of_mass(array, axis):
    """
    Compute the center of mass of an array over a given axis.

    Parameters
    ----------
    array : numpy.ndarray
        The input array.
    axis : int
        The axis on which to compute the center of mass.

    Returns
    -------
    float
        The position of the center of mass.
    """

    pos = np.arange(array.shape[axis])

    axes = list(range(array.ndim))
    axes.remove(axis)
    pos = np.expand_dims(pos, axes)

    return np.tensordot(pos, array, (axis, axis)).squeeze() / array.sum(axis)


def make_broadcastable(array, target):
    """
    Append the required amount of empty dimensions to an array to make it
    broadcastable with respect to another.

    Parameters
    ----------
    array : numpy.ndarray
        The array to be broadcasted.
    target : numpy.ndarray
        The target array, reference for the broadcast size.

    Returns
    -------
    np.ndarray
        The broadcastable array.

    Raises
    ------
    ValueError
        If the target array has fewer dimensions than the source.
    """

    if array.ndim > target.ndim:
        raise ValueError("Target array must have equal or more dimensions.")

    broadcast_shape = np.ones(target.ndim, dtype=int)
    for ax, dim in enumerate(array.shape):
        broadcast_shape[ax] = dim
    return array.reshape(broadcast_shape)


def transpose_dim_to(array, src_dim, target_dim):
    """
    Transpose the array so that the given dimension moves to the specified
    position and the order of the rest is unchanged.
    This is a convenience wrapper around numpy.transpose.

    Inversing the transpose is done as such::

        >>> N1, N2 = 1, 3
        >>> a0 = numpy.ones([2, 3, 4, 5])
        >>> a1 = transpose_dim_to(a0, N1, N2)
        >>> a2 = transpose_dim_to(a1, N2, N1)
        >>> a0 == a2


    Parameters
    ----------
    array : numpy.ndarray
        The array to be transposed.
    src_dim : int
        The dimension to be moved.
    target_dim : int
        The target position of the dimension.

    Returns
    -------
    numpy.ndarray
        The transposed array.
    """

    dims = list(range(array.ndim))
    dims.remove(src_dim)
    dims.insert(target_dim, src_dim)
    return np.transpose(array, dims)


def threshold_crop(array, threshold, axis, summed=False):
    """
    Crop an array to a relative threshold on given axis. The returned array is
    the smallest array that contains all elements superior to the threshold.
    The thresholding can be done per element, or based on the sum over the
    non cropping axes.

    Parameters
    ----------
    array : numpy.ndarray
        The input array.
    threshold : float
        Relative threshold for cropping.
    axis : int
        Axis along which to crop.
    summed : bool, optional
        Compute the summed thresholds, by default False

    Returns
    -------
    numpy.ndaray
        The cropped array.
    """

    dims = list(range(array.ndim))
    dims.remove(axis)
    dims = tuple(dims)

    if summed:
        amplitude = array.sum(dims)
    else:
        amplitude = array.max(dims)

    amplitude = amplitude / amplitude.max()
    thresh = np.nonzero(amplitude > threshold)[0]

    slices = [slice(None)] * array.ndim
    slices[axis] = slice(thresh[0], thresh[-1]+1)
    slices = tuple(slices)
    return array[slices]


def positive_index(index, size):
    """
    Return a positive index from any index. If the index is positive, it is
    returned. If negative, size+index will be returned.

    Parameters
    ----------
    index : int
        The input index.
    size : int
        The size of the indexed dimension.

    Returns
    -------
    int
        A positive index corresponding to the input index.

    Raises
    ------
    ValueError
        If the given index is not valid for the given size.
    """

    if not -size <= index <= size:
        raise ValueError(
            "Invalid index {} for size {}".format(index, size))

    if index < 0:
        index = size + index

    return index



# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import unittest
import numpy as np

import cbi_toolbox.splineradon as spl


class TestSplradon(unittest.TestCase):

    def test_dimension(self):
        image_dim = (50, 50, 3)
        image_3d = np.random.default_rng().random(image_dim)
        image_2d = np.copy(image_3d[..., 0])

        spline_3d = spl.radon(image_3d)
        spline_2d = spl.radon(image_2d)

        np.testing.assert_allclose(
            spline_2d, spline_3d[..., 0], rtol=1e-12, atol=1e-12)

        out_3d = spl.iradon(image_3d)
        out_2d = spl.iradon(image_2d)

        np.testing.assert_allclose(
            out_2d, out_3d[..., 0], rtol=1e-12, atol=1e-12)

    def test_contiguous(self):
        image_dim = (50, 50, 3)
        image_3d = np.random.default_rng().random(image_dim)

        image_3d = np.transpose(image_3d, (1, 0, 2))
        image_2d = np.copy(image_3d[..., 0])

        spline_3d = spl.radon(image_3d)
        spline_2d = spl.radon(image_2d)

        np.testing.assert_allclose(
            spline_2d, spline_3d[..., 0], rtol=1e-9, atol=1e-12)

        out_3d = spl.iradon(spline_3d)
        out_2d = spl.iradon(spline_2d)

        np.testing.assert_allclose(
            out_2d, out_3d[..., 0], rtol=1e-9, atol=1e-12)

    def test_padding(self):
        theta = np.arange(10)
        for size in range(5, 25):
            shape = (size, size)
            image = np.random.default_rng().random(shape)

            for circle in (True, False):
                rd = spl.radon(image, theta, circle=circle,
                               b_spline_deg=(0, 0))
                ird = spl.iradon(rd, theta, circle=circle, b_spline_deg=(0, 0))

                np.testing.assert_array_equal(shape, ird.shape)


if __name__ == "__main__":
    unittest.main()



"""
This module implements sinogram filtering for the FBP algorithm.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
# and Michael Liebling <michael.liebling@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is inspired from `SPLRADON` written by Michael Liebling
# <http://sybil.ece.ucsb.edu/pages/splineradon/splineradon.html>


import numpy as np
from scipy import fft

from cbi_toolbox.utils import make_broadcastable, fft_size


def get_filter(n, filter_name: str, degree=None):
    """
    Compute the frequency coefficients of the filter for the FBP.
    Only the positive frequencies are computed for use with RFFT/IRFFT (scipy).

    Parameters
    ----------
    n : int
        Size of the filter.
    filter_name : str
        Type of filter, one of: ['None', 'Ram-Lak', 'Shepp-Logan', 'Cosine'].
    degree : int, optional
        Degree of the filter, when applicable, by default None

    Returns
    -------
    numpy.ndarray
        The frequency coefficients of the FBP filter.

    Raises
    ------
    ValueError
        If the type of filter asked is not known.
    """

    filter_name = filter_name.upper()
    pre_filter = True

    freq = np.concatenate((np.arange(1, n / 2 + 1, 2, dtype=int),
                           np.arange(n / 2 - 1, 0, -2, dtype=int)))

    filter_f = np.zeros(n)
    filter_f[0] = 0.25
    filter_f[1::2] = -1 / (np.pi * freq) ** 2
    filter_f = 2 * np.real(fft.rfft(filter_f))

    if filter_name == 'NONE':
        filter_f = np.ones(filter_f.shape)

    elif filter_name == 'RAM-LAK':
        pass

    elif filter_name == 'SHEPP-LOGAN':
        omega = np.pi * fft.rfftfreq(n)[1:]
        filter_f[1:] *= np.sin(omega) / omega

    elif filter_name == 'COSINE':
        freq = np.linspace(np.pi/2, np.pi, filter_f.size, endpoint=False)
        cosine_filter = np.sin(freq)
        filter_f *= cosine_filter

    else:
        nu = np.arange((n + 1) / 2)
        n0 = 3

        k = 50
        k_vector = np.arange(-k, k + 0.5)
        k_vector = k_vector[..., np.newaxis]
        pre_filter = False

        if filter_name == 'B-SPLINE':
            filter_f = np.abs(
                nu) / np.sum(np.power(np.sinc(nu + k_vector), degree + 1), 0)

        elif filter_name == 'OBLIQUE':
            filter_f = np.abs(nu) / np.power(np.sinc(nu), degree + 1)

        elif filter_name == 'FRACTIONAL':
            filter_f = np.abs(np.sin(np.pi * nu) / np.pi) / \
                np.sum(np.power(np.abs(np.sinc(nu + k_vector)), degree + 2), 0)

        elif filter_name == 'FRACTIONALOBLIQUE':
            filter_f = np.abs(np.sin(np.pi * nu) / 2) / \
                np.power(np.abs(np.sinc(nu)), degree + 2)

        elif filter_name == 'BSPLINEPROJ':
            nu_k = nu + k_vector
            filter_f = np.sum(np.abs(nu_k) * np.power(np.sinc(nu_k), degree + 2 + n0), 0) / (
                np.sum(np.power(np.sinc(nu_k), n0 + 1), 0) *
                np.sum(np.power(np.abs(np.power(np.sinc(nu_k), degree + 1)), 2), 0))

        else:
            raise ValueError('Illegal filter name: {}'.format(filter_name))

    return filter_f, pre_filter


def filter_sinogram(sinogram, filter_name, degree=None):
    """
    Filter a sinogram for FBP reconstruction.

    Parameters
    ----------
    sinogram : numpy.ndarray
        The raw sinogram.
    filter_name : str
        Type of filter, one of: ['None', 'Ram-Lak', 'Shepp-Logan', 'Cosine'].
    degree : int, optional
        Degree of the filter, when applicable, by default None

    Returns
    -------
    numpy.ndarray
        The filtered sinogram.
    """

    length = sinogram.shape[1]
    n = fft_size(length)

    filter_f, pre_filter = get_filter(n, filter_name, degree)
    filtered = fft.rfft(sinogram, n, axis=1)
    filter_f = make_broadcastable(filter_f[np.newaxis, ...], filtered)
    filtered *= filter_f
    filtered = fft.irfft(filtered, axis=1, overwrite_x=True)

    return filtered[:, :length], pre_filter


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    n = 256

    filters = [
        'None',
        'Ram-Lak',
        'Shepp-Logan',
        'Cosine'
    ]

    freq = fft.rfftfreq(n)
    for filt in filters:
        f, _ = get_filter(n, filt)
        plt.plot(freq, f, label=filt)

    plt.legend()
    plt.show()



"""
The splineradon package implements radon and inverse radon transforms using
b-spline interpolation formulas described in [1] using multithreading and GPU
acceleration.

**Conventions:**

arrays follow the ZXY convention, with

    - Z : depth axis (axial, focus axis)
    - X : horizontal axis (lateral)
    - Y : vertical axis (lateral, rotation axis when relevant)

sinograms follow the TPY convention, with

    - T : angles (theta)
    - P : captor axis
    - Y : rotation axis

[1] S. Horbelt, M. Liebling, M. Unser, *"Discretization of the Radon Transform
and of Its Inverse by Spline Convolutions"*, IEEE Transactions on Medical Imaging,
vol 21, no 4, pp. 363-376, April 2002.


"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
# and Michael Liebling <michael.liebling@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is inspired from `SPLRADON` written by Michael Liebling
# <http://sybil.ece.ucsb.edu/pages/splineradon/splineradon.html>

import numpy as np

from cbi_toolbox.bsplines import change_basis
import cbi_toolbox.splineradon._cradon as cradon
from ._filter_sinogram import *
from ._spline_kernel import *


def is_cuda_available(verbose=False):
    """
    Check if CUDA can be used for computations.

    Parameters
    ----------
    verbose : bool, optional
        Print verbose errors, by default False.

    Returns
    -------
    bool
        Whether CUDA is available.
    """

    try:
        return cradon.is_cuda_available()
    except RuntimeError as e:
        if verbose:
            print(e)
        return False


def radon(image, theta=None, angledeg=True, n=None,
          b_spline_deg=(2, 3), sampling_steps=(1, 1),
          center=None, captors_center=None, circle=False,
          nt=200, use_cuda=False):
    """
    Perform a radon transform on the image.

    Parameters
    ----------
    image : numpy.ndarray [ZXY] or [ZX]
        The input image.
    theta : array_like, optional
        The sinogram angles, by default None.
        If None, uses numpy.arange(180).
    angledeg : bool, optional
        Give angles in degrees instead of radians, by default True.
    n : int, optional
        The number of captors (overrides sampling_steps[1]), by default None.
    b_spline_deg : tuple, optional
        (ni, ns) the degrees of the image and sinogram bspline bases,
        by default (2, 3).
    sampling_steps : tuple, optional
        Pixel sampling steps on the image and the sinogram, by default (1, 1).
    center : tuple, optional
        (z, x) the center of rotation of the image, by default None (centered).
    captors_center : float, optional
        The position of the center of rotation in the sinogram, by default None.
    circle : bool, optional
        If the object is contained in the inner circle/cylinder (will produce a
        sinogram with same width as the image), by default False.
    nt : int, optional
        Number of points stored in the spline kernel, by default 200.
    use_cuda : bool, optional
        Use CUDA GPU acceleration, by default False.

    Returns
    -------
    numpy.ndarray [TPY]
        The computed sinogram.
    """

    if theta is None:
        theta = np.arange(180)

    spline_image = radon_pre(image, b_spline_deg[0])
    sinogram = radon_inner(spline_image, theta, angledeg, n, b_spline_deg, sampling_steps,
                           center, captors_center, circle, nt, use_cuda=use_cuda)

    sinogram = radon_post(sinogram, b_spline_deg[1])

    return sinogram


def iradon(sinogram, theta=None, angledeg=True, filter_type='RAM-LAK',
           b_spline_deg=(2, 3), sampling_steps=(1, 1),
           center=None, captors_center=None, circle=False,
           nt=200, use_cuda=False):
    """
    Perform a filtered back projection on the sinogram.

    Parameters
    ----------
    sinogram : numpy.ndarray [TPY]
        The input sinogram.
    theta : array_like, optional
        The sinogram angles, by default None.
        If None, uses numpy.arange(180).
    angledeg : bool, optional
        Give angles in degrees instead of radians, by default True.
    filter_type : str, optional
        The type of filter used for FBP, by default 'RAM-LAK'.
        Can be one of ['None', 'Ram-Lak', 'Shepp-Logan', 'Cosine'].
    b_spline_deg : tuple, optional
        (ni, ns) the degrees of the image and sinogram bspline bases,
        by default (2, 3).
    sampling_steps : tuple, optional
        Pixel sampling steps on the image and the sinogram, by default (1, 1).
    center : tuple, optional
        (z, x) the center of rotation of the image, by default None (centered).
    captors_center : float, optional
        The position of the center of rotation in the sinogram, by default None.
    circle : bool, optional
        If the object is contained in the inner circle/cylinder (will produce a
        sinogram with same width as the image), by default False.
    nt : int, optional
        Number of points stored in the spline kernel, by default 200.
    use_cuda : bool, optional
        Use CUDA GPU acceleration, by default False.

    Returns
    -------
    numpy.ndarray [ZXY]
        The reconstructed image.
    """

    sinogram = iradon_pre(sinogram, b_spline_deg[1], filter_type, circle)

    image, theta = iradon_inner(sinogram, theta, angledeg, b_spline_deg, sampling_steps,
                                center, captors_center, nt, use_cuda=use_cuda)
    image = iradon_post(image, theta, b_spline_deg[0])

    return image


def radon_pre(image, ni=2):
    """
    Pre-processing step for radon transform.
    Projects the image onto a bspline basis.

    Parameters
    ----------
    image : numpy.ndarray [ZXY] or [ZX]
        The raw input image.
    ni : tuple, optional
        The degree of the image bspline basis, by default 2.

    Returns
    -------
    np.ndarray
        The projected image.
    """

    return change_basis(image, 'cardinal', 'b-spline', ni, (0, 1),
                        boundary_condition='periodic')


def radon_inner(spline_image, theta=None, angledeg=True, n=None,
                b_spline_deg=(2, 3), sampling_steps=(1, 1),
                center=None, captors_center=None, circle=False,
                nt=200, use_cuda=False):
    """
    Raw radon transform, requires pre and post-processing (projections onto
    bspline bases). This can be run in parallel by splitting theta.

    Parameters
    ----------
    spline_image : numpy.ndarray
        The input image in a bspline basis.
    theta : array_like, optional
        The sinogram angles, by default None.
        If None, uses numpy.arange(180).
    angledeg : bool, optional
        Give angles in degrees instead of radians, by default True.
    n : int, optional
        The number of captors (overrides sampling_steps[1]), by default None.
    b_spline_deg : tuple, optional
        (ni, ns) the degrees of the image and sinogram bspline bases,
        by default (2, 3).
    sampling_steps : tuple, optional
        Pixel sampling steps on the image and the sinogram, by default (1, 1).
    center : tuple, optional
        (z, x) the center of rotation of the image, by default None (centered).
    captors_center : float, optional
        The position of the center of rotation in the sinogram, by default None.
    circle : bool, optional
        If the object is contained in the inner circle/cylinder (will produce a
        sinogram with same width as the image), by default False.
    nt : int, optional
        Number of points stored in the spline kernel, by default 200.
    use_cuda : bool, optional
        Use CUDA GPU acceleration, by default False.

    Returns
    -------
    numpy.ndarray [TPY]
        The computed sinogram in a bspline basis.
    """

    if theta is None:
        theta = np.arange(180)

    nz = spline_image.shape[0]
    nx = spline_image.shape[1]
    h = sampling_steps[0]
    s = sampling_steps[1]
    ni = b_spline_deg[0]
    ns = b_spline_deg[1]

    shape = np.max(spline_image.shape[0:2])
    if not circle:
        nc = int(np.ceil(shape * np.sqrt(2)))
        nc += (nc % 2 != shape % 2)

    else:
        nc = shape

    if n is not None:
        s = (nc - 1) / (n - 1)
        nc = n

    if center is None:
        center = (nz - 1) / 2, (nx - 1) / 2

    if captors_center is None:
        captors_center = s * (nc - 1) / 2

    if angledeg:
        theta = np.deg2rad(theta)

    kernel = get_kernel_table(nt, ni, ns, h, s, -theta, degree=False)

    squeeze = False
    if spline_image.ndim < 3:
        spline_image = spline_image[..., np.newaxis]
        squeeze = True

    sinogram = cradon.radon(
        spline_image,
        h,
        ni,
        center[0],
        center[1],
        -theta,
        kernel[0],
        kernel[1],
        nc,
        s,
        ns,
        captors_center,
        use_cuda
    )

    if squeeze:
        sinogram = np.squeeze(sinogram)

    return sinogram


def radon_post(sinogram, ns=3):
    """
    Post-processing for the radon transform.
    Projects the sinogram back from a bspline basis.

    Parameters
    ----------
    sinogram : numpy.ndarray
        The sinogram in bspline basis.
    ns : tuple, optional
        The degree of the sinogram bspline basis, by default 3

    Returns
    -------
    numpy.ndarray
        The sinogram.
    """

    if ns > -1:
        sinogram = change_basis(sinogram, 'dual', 'cardinal',
                                ns, 1, boundary_condition='periodic', in_place=True)
    return sinogram


def iradon_pre(sinogram, ns=3, filter_type='RAM-LAK', circle=False):
    """
    Pre-processing for the inverse radon transform.
    Filters the sinogram and projects it onto a bspline basis.

    Parameters
    ----------
    sinogram : numpy.ndarray
        The raw sinogram.
    ns : tuple, optional
        Degree of the sinogram bspline basis, by default 3.
    filter_type : str, optional
        The type of filter used for FBP, by default 'RAM-LAK'.
        Can be one of ['None', 'Ram-Lak', 'Shepp-Logan', 'Cosine'].
    circle : bool, optional
        If the object is contained in the inner circle/cylinder (will produce a
        sinogram with same width as the image), by default False.

    Returns
    -------
    numpy.ndarray
        The filtered and projected sinogram.
    """

    if circle:
        shape = sinogram.shape[1]
        nc = np.ceil(shape * np.sqrt(2))
        nc += (nc % 2 != shape % 2)
        pad = int((nc - shape) // 2)
        padding = [(0, )] * sinogram.ndim
        padding[1] = (pad, )
        sinogram = np.pad(sinogram, padding)

    sinogram, pre_filter = filter_sinogram(sinogram, filter_type, ns)

    if pre_filter:
        sinogram = change_basis(sinogram, 'CARDINAL', 'B-SPLINE', ns, 1,
                                boundary_condition='periodic')
    return sinogram


def iradon_inner(sinogram_filtered, theta=None, angledeg=True,
                 b_spline_deg=(2, 3), sampling_steps=(1, 1),
                 center=None, captors_center=None,
                 nt=200, use_cuda=False):
    """
    Raw inverse radon transform, requires pre and post-processing for sinogram
    filtering and change to bspline basis.
    Can be run in parallel by splitting the sinogram and theta.

    Parameters
    ----------
    sinogram_filtered : numpy.ndarray [TPY]
        The pre-processed sinogram.
    theta : array_like, optional
        The sinogram angles, by default None.
        If None, uses numpy.arange(180).
    angledeg : bool, optional
        Give angles in degrees instead of radians, by default True.
    b_spline_deg : tuple, optional
        (ni, ns) the degrees of the image and sinogram bspline bases,
        by default (2, 3).
    sampling_steps : tuple, optional
        Pixel sampling steps on the image and the sinogram, by default (1, 1).
    center : tuple, optional
        (z, x) the center of rotation of the image, by default None (centered).
    captors_center : float, optional
        The position of the center of rotation in the sinogram, by default None.
    nt : int, optional
        Number of points stored in the spline kernel, by default 200.
    use_cuda : bool, optional
        Use CUDA GPU acceleration, by default False.

    Returns
    -------
    numpy.ndarray
        The image in bspline basis.
    """

    nc = sinogram_filtered.shape[1]
    na = sinogram_filtered.shape[0]

    ni = b_spline_deg[0]
    ns = b_spline_deg[1]
    h = sampling_steps[0]
    s = sampling_steps[1]

    if theta is None:
        theta = np.pi / na
        angledeg = False

    theta = np.atleast_1d(theta)

    if theta.size == 1 and na > 1:
        theta = np.arange(na) * theta

    if angledeg:
        theta = np.deg2rad(theta)

    kernel = get_kernel_table(nt, ni, ns, h, s, -theta, degree=False)

    nx = int(np.floor(nc / np.sqrt(2)))
    nx -= (nx % 2 != nc % 2)
    nz = nx

    if center is None:
        center = (nz - 1) / 2, (nx - 1) / 2

    if captors_center is None:
        captors_center = s * (nc - 1) / 2

    squeeze = False
    if sinogram_filtered.ndim < 3:
        sinogram_filtered = sinogram_filtered[..., np.newaxis]
        squeeze = True

    image = cradon.iradon(
        sinogram_filtered,
        s,
        ns,
        captors_center,
        -theta,
        kernel[0],
        kernel[1],
        nz,
        nx,
        h,
        ni,
        center[0],
        center[1],
        use_cuda
    )

    if squeeze:
        image = np.squeeze(image)

    return image, theta


def iradon_post(image, theta, ni=2):
    """
    Post-processing for the inverse Radon transform.
    Projects the image back from a bspline basis and normalizes it.

    Parameters
    ----------
    image : numpy.ndarray
        The image in bspline basis.
    theta : array_like
        The projection angles of the sinogram.
    ni : tuple, optional
        The degree of the image bspline basis, by default 2.

    Returns
    -------
    numpy.ndarray
        The reconstructed image.
    """

    if ni > -1:
        image = change_basis(image, 'DUAL', 'CARDINAL', ni,
                             (0, 1), boundary_condition='periodic', in_place=True)

    if theta.size > 1:
        image = image * np.pi / (2 * theta.size)

    return image



"""
The preprocess module implements functions to preprocess experimental images.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import cv2
import numpy as np


def erase_corners(image_array, corner_size=300):
    """
    Fill the corners of the image with the closest pixel value.
    This is useful when a diaphragm is visible in the field of view.

    Parameters
    ----------
    image_array : numpy.ndarray
        The original image.
    corner_size : int, optional
        The size of the corner to fill, by default 300

    Returns
    -------
    numpy.ndarray
        The new image with corners filled.
    """

    if corner_size > 0:
        image_array[..., :corner_size, :corner_size] = image_array[...,
                                                                   corner_size, corner_size, None, None]
        image_array[..., :corner_size, -corner_size:] = image_array[...,
                                                                    corner_size, -corner_size, None, None]
        image_array[..., -corner_size:, -corner_size:] = image_array[..., -
                                                                     corner_size, -corner_size, None, None]
        image_array[..., -corner_size:, :corner_size] = image_array[..., -
                                                                    corner_size, corner_size, None, None]

    return image_array


def transmission_to_absorption(image_array, max_value=4096):
    """
    Convert a transmission image into an absorption one.
    This inverses the black-white contrast.

    Parameters
    ----------
    image_array : numpy.ndarray
        The absorption image.
    max_value : float, optional
        The max value for scaling the original image, by default 4096.

    Returns
    -------
    numpy.ndarray
        The absorption contrast image.
    """

    scaled_array = image_array / max_value

    absorption = np.log(1 / scaled_array)

    return absorption


def remove_background_illumination(image_array, threshold=0.5, hole_size=250,
                                   margin_size=100, border_axis=-1):
    """
    Removes the background illumination from images using thresholding and
    morphological filtering.

    Parameters
    ----------
    image_array : numpy.ndarray
        The images to be processed as a 3D array, the first dimension iterates
        over the different images.
    threshold : float, optional
        The relative threshold used to detect background, by default 0.5.
    hole_size : int, optional
        Biggest holes removed by filtering, by default 250.
    margin_size : int, optional
        Margin kept around the useful information, by default 100.
    border_axis : int, optional
        Axis used to detect the outer edge of the image, by default -1.

    Returns
    -------
    numpy.ndarray
        The array of images without background illumination.
    """

    mask_bool = image_array < threshold
    mask_int = mask_bool.astype(np.uint8)

    kernel_open = np.ones((hole_size, hole_size), dtype=np.uint8)
    kernel_dilate = np.ones((margin_size, margin_size), dtype=np.uint8)

    if border_axis < 0:
        contour_border_axis = border_axis
    else:
        contour_border_axis = border_axis - 1

    for plane_idx, plane_mask in enumerate(mask_int):
        find_contours = cv2.findContours(
            plane_mask, mode=cv2.RETR_EXTERNAL, method=cv2.CHAIN_APPROX_NONE)

        major = cv2.__version__.split('.')[0]
        if major == '2':
            contours = find_contours[1]
        else:
            contours = find_contours[0]

        plane_mask.fill(0)

        for index, contour in enumerate(contours):
            if len(contour) > 500 and (
                    contour[..., contour_border_axis].max() == image_array.shape[border_axis] - 1 or (
                    contour[..., contour_border_axis].min() == 0)):
                temp_mask = np.zeros_like(mask_int[plane_idx, ...])
                cv2.drawContours(temp_mask, contours, index, 1, cv2.FILLED)

                temp_mask = cv2.morphologyEx(
                    temp_mask, cv2.MORPH_CLOSE, kernel_open)
                mask_int[plane_idx, ...] |= temp_mask

        mask_int[plane_idx, ...] = cv2.morphologyEx(
            mask_int[plane_idx, ...], cv2.MORPH_ERODE, kernel_dilate)

    mask_bool = mask_int.astype(bool)
    image_array[mask_bool] = 0

    return image_array



"""
The parallel package provides tools to split parallel computations.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import os
import concurrent.futures


def distribute_bin(dimension, rank, workers):
    """
    Computes the start index and bin size to evenly split array-like data into
    multiple bins.

    Parameters
    ----------
    dimension : int
        The size of the array to distribute.
    rank : int, optional
        The rank of the worker.
    workers : int, optional
        The total number of workers.

    Returns
    -------
    (int, int)
        The start index of this bin, and its size.
        The distributed data should be array[start:start + bin_size].
    """

    if workers > dimension:
        workers = dimension

    if rank >= workers:
        return 0, 0

    bin_size = dimension // workers
    large_bin_number = dimension - bin_size * workers

    bin_index = 0

    if rank < large_bin_number:
        bin_size += 1
    else:
        bin_index += large_bin_number

    bin_index += rank * bin_size

    return bin_index, bin_size


def distribute_bin_all(dimension, workers):
    """
    Computes the start indexes and bin sizes of all splits to distribute
    computations across multiple workers.

    Parameters
    ----------
    dimension : int
        the size of the array to be distributed
    workers : int, optional
        the amount of workers

    Returns
    -------
    ([int], [int])
        The list of start indexes and the list of bin sizes to distribute data.
    """

    original_size = workers
    if workers > dimension:
        workers = dimension

    bin_size = dimension // workers
    large_bin_number = dimension - bin_size * workers

    bin_index = 0
    bin_indexes = []
    bin_sizes = []

    for j_index in range(original_size):
        if j_index >= workers:
            bin_indexes.append(0)
            bin_sizes.append(0)
            continue

        l_bin_size = bin_size
        if j_index < large_bin_number:
            l_bin_size += 1

        bin_indexes.append(bin_index)
        bin_sizes.append(l_bin_size)

        bin_index += l_bin_size

    return bin_indexes, bin_sizes


def parallelize(func, size, workers=None):
    """
    Launches a function multiple times in parallel using multithreading.
    Useful only if the GIL is released in the parallelized function (this is
    the case for many ``numpy`` and ``scipy`` routines).

    Parameters
    ----------
    func : function (callable)
        The function that will be run in parallel. It must take 2 arguments,
        which are the returns of ``distribute_bin_all`` corresponding to the
        thread pool (the list of starting indexes of data bins, and the list of
        bin sizes).
    size : int
        The size of the array that will be split between workers.
    workers : int, optional
        The maximum number of workers, by default None (will be maximized for
        the system).

    Returns
    -------
    iterator
        An iterator containing the results of the function calls, in a random
        order (see concurrent.futures.ThreadPoolExecutor.map).
    """

    try:
        omp_threads = int(os.environ['OMP_NUM_THREADS'])
    except KeyError:
        omp_threads = os.cpu_count()

    uworkers = os.cpu_count() if workers is None else workers

    workers = min(omp_threads, len(
        os.sched_getaffinity(0)), size, uworkers)

    if workers == 1:
        out = func(0, size)
        outputs = [out]

    else:
        bins, bin_dims = distribute_bin_all(size, workers)

        with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
            outputs = executor.map(func, bins, bin_dims)

    return outputs



# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause


import unittest
import numpy as np
from scipy import signal

from cbi_toolbox import bsplines


def test_convert(sig, degree, tolerance, condition):
    source = sig.copy()
    coeffs = bsplines.convert_to_interpolation_coefficients(source, degree, tolerance,
                                                            boundary_condition=condition, in_place=True)
    r_signal = bsplines.convert_to_samples(
        coeffs, degree, condition, in_place=True)
    np.testing.assert_allclose(sig, r_signal)


def test_basis(data, from_b, to_b, degree, axes, tolerance, condition):
    source = data.copy()
    c_out = bsplines.change_basis(
        source, from_b, to_b, degree, axes, tolerance, condition, in_place=True)
    ar = bsplines.change_basis(
        c_out, to_b, from_b, degree, axes, tolerance, condition, in_place=True)

    np.testing.assert_allclose(data, ar, rtol=tolerance * 10, atol=1e-12)


def test_dims(data, from_b, to_b, degree, axes, tolerance, condition):
    source = data.copy()
    sourceb = data.copy()
    out_1 = bsplines.change_basis(
        source, from_b, to_b, degree, axes, tolerance, condition, in_place=True)
    out_2 = bsplines.change_basis(
        sourceb[..., 0], from_b, to_b, degree, axes, tolerance, condition, in_place=True)

    np.testing.assert_allclose(
        out_1[..., 0], out_2)


def test_dims_conversion(data, degree, tolerance, condition):
    source = data.copy()
    sourceb = data.copy()
    c_1 = bsplines.convert_to_interpolation_coefficients(source, degree, tolerance,
                                                         boundary_condition=condition, in_place=True)
    c_2 = bsplines.convert_to_interpolation_coefficients(sourceb[..., 0], degree, tolerance,
                                                         boundary_condition=condition, in_place=True)

    np.testing.assert_allclose(c_1[..., 0], c_2)


class TestBsplines(unittest.TestCase):
    def test_convert(self):
        tolerance = 1e-15
        data_size = 50
        condition_list = ['mirror', 'periodic']
        data_size_list = [(data_size,), (data_size, 4), (data_size, 2, 2)]

        for degree in range(8):
            for size in data_size_list:
                sig = np.random.default_rng().random(size)
                for condition in condition_list:
                    test_convert(sig, degree, tolerance, condition)

    def test_basis(self):
        tolerance = 1e-15
        data_size = 50
        condition_list = ['mirror',  'periodic']
        data_size_list = [(data_size, data_size), (data_size, data_size, 2)]
        axes_list = [(0,), (0, 1)]
        bases_list = [('cardinal', 'b-spline'), ('cardinal', 'dual')]

        for degree in range(4):
            for size in data_size_list:
                sig = np.random.default_rng().random(size)
                for condition in condition_list:
                    for axis in axes_list:
                        for base in bases_list:
                            from_b, to_b = base
                            test_basis(sig, from_b, to_b, degree,
                                       axis, tolerance, condition)

    def test_dims(self):
        tolerance = 1e-16
        data_size = 50
        condition_list = ['mirror', 'periodic']
        data_size_list = [(data_size, data_size, 2)]
        axes_list = [(0,), (0, 1)]
        bases_list = [('cardinal', 'b-spline'), ('cardinal', 'dual')]

        for degree in range(4):
            for size in data_size_list:
                sig = np.random.default_rng().random(size)
                for condition in condition_list:
                    test_dims_conversion(
                        sig, degree, tolerance, condition)
                    for axis in axes_list:
                        for base in bases_list:
                            from_b, to_b = base
                            test_dims(sig, from_b, to_b, degree,
                                      axis, tolerance, condition)

    def test_splines(self):
        x = np.linspace(-10, 10, 10000)
        for degree in range(9):
            np.testing.assert_allclose(signal.bspline(
                x, degree), bsplines.bspline(x, degree))


if __name__ == '__main__':
    unittest.main()



"""
The utils package provides various utility functions to work with files and
arrays.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import json
import os
import numpy as np
import apeer_ometiff_library.io as omeio

from cbi_toolbox.utils._arrays import *


def load_ome_tiff(file_path):
    """
    Load an OME tiff file as a numpy array [ZXY].

    Parameters
    ----------
    file_path : string
        The file to load.

    Returns
    -------
    array [ZXY]
        The loaded array.
    """

    array, xmlstring = omeio.read_ometiff(file_path)
    array = array.squeeze()
    array = np.ascontiguousarray(array.transpose((0, 2, 1)))

    return array, xmlstring


def save_ome_tiff(file_path, image, xmlstring=None):
    """
    Save numpy array to OME tiff format.

    Parameters
    ----------
    file_path : string
        Where to save the data.
    image : array[ZXY]
        The array to save.
    xmlstring : str, optional
        Xml metadata, by default None.
    """

    image = image.transpose((0, 2, 1))
    image = image[None, :, None, ...]
    omeio.write_ometiff(file_path, image, xmlstring)


def fft_size(n):
    """
    Returns the smallest power of 2 above n, but no less than 64
    for efficient FFT computations.

    Parameters
    ----------
    n : int
        Size of the signal.
    """

    return max(64, int(2 ** np.ceil(np.log2(n))))



"""
CBI Toolbox contains a collection of algorithms used for computational
bioimaging and microscopy.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause


from os import path
import os
import platform
import subprocess
import sys
import glob

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext

__version__ = '1.0.2'

requires = [
    'numpy',
    'opencv-python',
    'apeer-ometiff-library',
    'scipy>=1.6.0',
    'scikit-image',
    'noise',
    'poppy',
]

extras_require = {
    'plots': ['napari', 'matplotlib'],
    'mpi': ['mpi4py'],
    'docs': ['sphinx', 'sphinxcontrib-apidoc'],
}


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        self.sourcedir = os.path.abspath(sourcedir)

        sources = glob.glob('pybind11/tools/*')
        sources.extend(glob.glob('pybind11/include/**/*', recursive=True))
        sources.append('pybind11/CMakeLists.txt')
        sources.extend(glob.glob('**/src/*', recursive=True))
        sources.append('CMakeLists.txt')

        Extension.__init__(self, name, sources=sources)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            if isinstance(ext, CMakeExtension):
                self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(
            self.get_ext_fullpath(ext.name)))

        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg, '--clean-first']

        if platform.system() == "Windows":
            cmake_args += [
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2 ** 32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if os.path.exists(self.build_temp):
            import shutil
            shutil.rmtree(self.build_temp)

        os.makedirs(self.build_temp)

        subprocess.check_call(['cmake', ext.sourcedir] +
                              cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] +
                              build_args, cwd=self.build_temp)


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='cbi_toolbox',
    version=__version__,
    author='François Marelli',
    author_email='francois.marelli@idiap.ch',
    description='A python toolbox for computational bioimaging',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD-3',
    url='https://github.com/idiap/cbi_toolbox',
    ext_modules=[
        CMakeExtension('cbi_toolbox.cmake_ext'),
    ],
    packages=find_packages(exclude=('tests', )),
    install_requires=requires,
    python_requires='>=3.7',
    extras_require=extras_require,
    cmdclass={'build_ext': CMakeBuild},
    zip_safe=False,
)



"""
The reconstruct package provides reconstruction algorithms,
as well as preprocessing tools and performance scores.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np


def psnr(ref, target, norm=None, limit=None):
    """
    Computes the Peak Signal-to-Noise Ratio:
    PSNR = 10 log( limit ^ 2 / MSE(ref, target) )

    Parameters
    ----------
    ref : numpy.ndarray
        The ground-truth reference array.
    target : numpy.ndarray
        The reconstructed array.
    norm : str
        Normalize the images before computing snr, default is None.
    limit: float, optional
        The maximum pixel value used for PSNR computation,
        default is None (max(ref)).

    Returns
    -------
    float
        The PSNR.
    """

    if norm is None:
        pass
    elif norm == 'mse':
        scale_to_mse(ref, target)
    else:
        normalize(ref)
        normalize(target)

    if limit is None:
        limit = ref.max()

    return 10 * np.log10(limit**2 / mse(ref, target))


def mse(ref, target):
    """
    Computes the Mean Squared Error between two arrays

    Parameters
    ----------
    ref : numpy.ndarray
        Reference array.
    target : numpy.ndarray
        Target array.

    Returns
    -------
    float
        The MSE.
    """

    return np.square(np.subtract(ref, target)).mean()


def normalize(image, mode='std'):
    """
    Normalize an image according to the given criterion.

    Parameters
    ----------
    image : numpy.ndarray
        Image to normalize, will be modified.
    mode : str, optional
        Type of normalization to use, by default 'std'.
        Allowed: ['std', 'max', 'sum']

    Returns
    -------
    array
        The normalized image (same as input).

    Raises
    ------
    ValueError
        For unknown mode.
    """

    if mode == 'std':
        f = np.std(image)
    elif mode == 'max':
        f = np.max(image)
    elif mode == 'sum':
        f = np.sum(image)
    else:
        raise ValueError('Invalid norm: {}'.format(mode))

    image /= f
    return image


def scale_to_mse(ref, target):
    """
    Scale a target array to minimise MSE with reference

    Parameters
    ----------
    ref : numpy.ndarray
        The reference for MSE.
    target : numpy.ndarray
        The array to rescale (will be done in-place).

    Returns
    -------
    numpy.ndarray
        The rescaled target.
    """
    w = np.sum(ref * target) / np.sum(target ** 2)
    target *= w

    return target



"""
The fpsopt module implements algorithms to deconvolve sinograms acquired using
Focal-Plane-Scanning OPT [1].

**Conventions:**

arrays follow the ZXY convention, with

    - Z : depth axis (axial, focus axis)
    - X : horizontal axis (lateral)
    - Y : vertical axis (lateral, rotation axis when relevant)

sinograms follow the TPY convention, with

    - T : angles (theta)
    - P : captor axis
    - Y : rotation axis

[1] K. G. Chan and M. Liebling, *"Direct inversion algorithm for focal plane
scanning optical projection tomography"*, in Biomedical Optics Express, vol. 8,
no. 11, pp. 5349-5358, 2017.
"""

# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause


# This code is derived from `fpsopt`, which includes the following license:
#
#  Modified BSD-2 License - for Non-Commercial Research and Educational Use Only
#
# Copyright (c) 2017, The Regents of the University of California
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for non-commercial research and educational use
# only provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# For permission to use for commercial purposes, please contact UCSB's Office of
# Technology & Industry Alliances at 805-893-5180 or info@tia.ucsb.edu.

from scipy import fft
import numpy as np
from cbi_toolbox.utils import fft_size


def inverse_psf_rfft(psf, shape=None, l=20, mode='laplacian'):
    """
    Computes the real FFT of a regularized inversed 2D PSF (or projected 3D)
    This follows the convention of fft.rfft: only half the spectrum is computed.

    Parameters
    ----------
    psf : array [ZXY] or [XY]
        The 2D PSF (if 3D, will be projected on Z axis).
    shape : tuple (int, int), optional
        Shape of the full-sized desired PSF
        (if None, will be the same as the PSF), by default None.
    l : int, optional
        Regularization lambda, by default 20
    mode : str, optional
        The regularizer used, by default laplacian.
        One of: ['laplacian', 'constant']

    Returns
    -------
    array [XY]
        The real FFT of the inverse PSF.

    Raises
    ------
    ValueError
        If the PSF has incorrect number of dimensions.
        If the regularizer is unknown.
    """

    if psf.ndim == 3:
        psf = psf.sum(0)
    elif psf.ndim != 2:
        raise ValueError("Invalid dimensions for PSF: {}".format(psf.ndim))

    if shape is None:
        shape = psf.shape

    psf_fft = fft.rfft2(psf, s=shape)

    # We need to shift the PSF so that the center is located at the (0, 0) pixel
    # otherwise deconvolving will shift every pixel
    freq = fft.rfftfreq(shape[1])
    phase_shift = freq * 2 * np.pi * ((psf.shape[1] - 1) // 2)
    psf_fft *= np.exp(1j * phase_shift[None, :])

    freq = fft.fftfreq(shape[0])
    phase_shift = freq * 2 * np.pi * ((psf.shape[0] - 1) // 2)
    psf_fft *= np.exp(1j * phase_shift[:, None])

    if mode == 'laplacian':
        # Laplacian regularization filter to avoid NaN
        filt = [[0, -0.25, 0], [-0.25, 1, -0.25], [0, -0.25, 0]]
        reg = np.abs(fft.rfft2(filt, s=shape)) ** 2
    elif mode == 'constant':
        reg = 1
    else:
        raise ValueError('Unknown regularizer: {}'.format(mode))

    return psf_fft.conjugate() / (np.abs(psf_fft) ** 2 + l * reg)


def deconvolve_sinogram(sinogram, psf, l=20, mode='laplacian', clip=True):
    """
    Deconvolve a sinogram with given PSF.

    Parameters
    ----------
    sinogram : numpy.ndarray [TPY]
        The blurred sinogram.
    psf : numpy.ndarray [ZXY] or [XY]
        The PSF used to deconvolve.
    l : float, optional
        Strength of the regularization.
    clip : bool, optional
        Clip negative values to 0, default is True.

    Returns
    -------
    numpy.ndarray [TPY]
        The deconvolved sinogram.
    """

    fft_shape = [fft_size(s) for s in sinogram.shape[1:]]

    inverse = inverse_psf_rfft(psf, shape=fft_shape, l=l, mode=mode)

    s_fft = fft.rfft2(sinogram, s=fft_shape)
    i_fft = fft.irfft2(s_fft * inverse, s=fft_shape, overwrite_x=True)
    i_fft = i_fft[:, :sinogram.shape[1], :sinogram.shape[2]]
    if clip:
        np.clip(i_fft, 0, None, out=i_fft)

    return i_fft


if __name__ == '__main__':
    from cbi_toolbox.simu import optics, primitives, imaging
    import cbi_toolbox.splineradon as spl
    import napari

    TEST_SIZE = 64

    s_psf = optics.gaussian_psf(
        numerical_aperture=0.3,
        npix_axial=TEST_SIZE+1, npix_lateral=TEST_SIZE+1)

    i_psf = inverse_psf_rfft(s_psf, l=1e-15, mode='constant')
    psfft = fft.rfft2(s_psf.sum(0))
    dirac = fft.irfft2(psfft * i_psf, s=s_psf.shape[1:])

    sample = primitives.boccia(
        TEST_SIZE, radius=(0.8 * TEST_SIZE) // 2, n_stripes=4)
    s_theta = np.arange(90)

    s_radon = spl.radon(sample, theta=s_theta, circle=True)
    s_fpsopt = imaging.fps_opt(sample, s_psf, theta=s_theta)

    s_deconv = deconvolve_sinogram(s_fpsopt, s_psf, l=0)

    print(np.square(s_deconv - s_radon).max()**0.5)

    with napari.gui_qt():
        viewer = napari.view_image(s_radon)
        viewer.add_image(s_fpsopt)
        viewer.add_image(s_deconv)

        viewer = napari.view_image(fft.fftshift(
            np.abs(i_psf), 0), name='inverse PSF FFT')
        viewer.add_image(dirac)



# Copyright (c) 2020 Idiap Research Institute, http://www.idiap.ch/
# Written by François Marelli <francois.marelli@idiap.ch>
#
# This file is part of CBI Toolbox.
#
# CBI Toolbox is free software: you can redistribute it and/or modify
# it under the terms of the 3-Clause BSD License.
#
# CBI Toolbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# 3-Clause BSD License for more details.
#
# You should have received a copy of the 3-Clause BSD License along
# with CBI Toolbox. If not, see https://opensource.org/licenses/BSD-3-Clause.
#
# SPDX-License-Identifier: BSD-3-Clause


import unittest
import numpy as np
from scipy import fft

from cbi_toolbox.simu import primitives, imaging, optics
from cbi_toolbox.reconstruct import fpsopt, psnr
from cbi_toolbox import splineradon as spl


class TestDeconv(unittest.TestCase):
    def setUp(self):
        psf_sizes = [65, 66]
        self.psfs = []

        for s in psf_sizes:
            self.psfs.append(optics.gaussian_psf(
                numerical_aperture=0.3,
                npix_axial=s, npix_lateral=s))

    def test_dirac(self):
        for psf in self.psfs:
            i_psf = fpsopt.inverse_psf_rfft(psf, l=0)
            psfft = fft.rfft2(psf.sum(0))
            dirac = fft.irfft2(psfft * i_psf, s=psf.shape[1:])

            ref = np.zeros_like(dirac)
            ref[(psf.shape[0] - 1)//2, (psf.shape[0] - 1)//2] = 1

            np.testing.assert_allclose(dirac, ref, atol=1e-12)

    def test_deconv(self):
        size = 100
        sample = primitives.boccia(
            size, radius=(0.5 * size) // 2, n_stripes=4)
        theta = np.arange(90)

        radon = spl.radon(sample, theta=theta, circle=True)

        for psf in self.psfs:
            s_fpsopt = imaging.fps_opt(sample, psf, theta=theta)
            decon = fpsopt.deconvolve_sinogram(s_fpsopt, psf, l=1e-12)
            snr = psnr(radon, decon)

            self.assertGreater(snr, 60)


if __name__ == "__main__":
    unittest.main()
